from typing import Callable, Dict, List, Optional, Tuple

import cv2
import numpy as np
from muscima.io import CropObject

from smashcima.geometry import Point, Vector2
from smashcima.scene import ComposedGlyph, Glyph, LineGlyph, ScenePoint, Sprite
from smashcima.scene.AffineSpace import AffineSpace
from smashcima.scene.SmashcimaLabels import SmashcimaLabels
from smashcima.scene.SmuflLabels import SmuflLabels

from .get_line_endpoints import get_line_endpoints
from .MppGlyphMetadata import MppGlyphMetadata
from .MppPage import MppPage

# source:
# https://pages.cvc.uab.es/cvcmuscima/index_database.html
MUSCIMA_PP_DPI = 300

TALL_BARLINE_THRESHOLD_PX = 150
BEAM_HOOK_MAX_WIDTH_PX = 25


def _mpp_mask_to_sprite_bitmap(mask: np.ndarray):
    """True/False pixel mask to black on transparent BGRA uint8 bitmap"""
    assert len(mask.shape) == 2
    assert mask.dtype == np.uint8
    alpha = mask * 255
    color = np.zeros_like(mask)
    bitmap = np.stack([color, color, color, alpha], axis=2)
    return bitmap


def _crop_objects_to_single_sprite_glyphs(
    crop_objects: List[CropObject],
    page: MppPage,
    label: str,
    sprite_origin: Optional[Callable[[CropObject], Point]] = None
) -> List[Glyph]:
    glyphs: List[Glyph] = []

    for o in crop_objects:
        space = AffineSpace()
        sprite = Sprite(
            space=space,
            bitmap=_mpp_mask_to_sprite_bitmap(o.mask),
            bitmap_origin=(
                sprite_origin(o) if sprite_origin else Point(0.5, 0.5)
            ),
            dpi=MUSCIMA_PP_DPI
        )
        glyph = Glyph(
            space=space,
            region=Glyph.build_region_from_sprites_alpha_channel(
                label=label,
                sprites=[sprite]
            ),
            sprites=[sprite]
        )
        MppGlyphMetadata.stamp_glyph(glyph, page, int(o.objid))
        glyphs.append(glyph)

    return glyphs


def _get_y_position_of_staff_line(
    page: MppPage,
    obj: CropObject,
    line_from_top: int = 0
) -> int:
    """
    Given a CropObject it finds the y-coordinate of the corresponding staff line
    """
    staff = page.get_outlink_to(obj, "staff")
    staff_line = None
    line = 0
    for l in staff.outlinks:
        resolved_link = page.id_lookup[l]
        if resolved_link.clsname == "staff_line":
            if line == line_from_top:  # counted from top, from zero
                staff_line = resolved_link
                break
            line += 1
    assert staff_line is not None
    return (staff_line.top + staff_line.bottom) // 2


def _get_symbols_centered_on_line(
    page: MppPage,
    clsname: str,
    label: str,
    line_from_top: int,
    when_center_outside_recenter: bool = False
) -> List[Glyph]:
    """
    Returns list of symbols with given clsname centered on given line index
    """
    def _sprite_origin(obj: CropObject) -> Point:
        line_y = _get_y_position_of_staff_line(
            page,
            obj,
            line_from_top=line_from_top
        )
        origin_y = (line_y - obj.top) / obj.height
        if (origin_y < 0 or origin_y > 1) and when_center_outside_recenter:
            origin_y = 0.5
        return Point(0.5, origin_y)

    return _crop_objects_to_single_sprite_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname == clsname
        ],
        page=page,
        label=label,
        sprite_origin=_sprite_origin
    )


def _crop_objects_to_line_glyphs(
    crop_objects: List[CropObject],
    page: MppPage,
    label: str,
    horizontal_line: bool,
    in_increasing_direction: bool,
) -> List[LineGlyph]:
    glyphs: List[LineGlyph] = []

    for o in crop_objects:
        # extract endpoints
        blurred_mask = cv2.medianBlur(o.mask, 5) # smooth out (5x5 window)
        points = get_line_endpoints(blurred_mask)
        points.sort(
            key=lambda p: p.x if horizontal_line else p.y,
            reverse=not in_increasing_direction
        )
        if len(points) < 2:
            # print(
            #     "Skipping line:", o.uid,
            #     "Has points:", len(points),
            #     "Is:", o.clsname
            # )
            continue

        # store the points in the point cloud
        page.point_cloud.set_points(o, [points[0], points[-1]])

        # construct the glyph
        space = AffineSpace()
        sprite = Sprite(
            space=space,
            bitmap=_mpp_mask_to_sprite_bitmap(o.mask),
            bitmap_origin=Point(0.5, 0.5),
            dpi=MUSCIMA_PP_DPI
        )
        start_point = ScenePoint(
            point=sprite.get_pixels_to_origin_space_transform().apply_to(points[0]),
            space=space
        )
        end_point = ScenePoint(
            point=sprite.get_pixels_to_origin_space_transform().apply_to(points[-1]),
            space=space
        )
        glyph = LineGlyph(
            space=space,
            region=Glyph.build_region_from_sprites_alpha_channel(
                label=label,
                sprites=[sprite]
            ),
            sprites=[sprite],
            start_point=start_point,
            end_point=end_point
        )
        MppGlyphMetadata.stamp_glyph(glyph, page, int(o.objid))

        # return the glyph
        glyphs.append(glyph)

    return glyphs


def get_connected_components_not_touching_image_border(
    mask: np.ndarray
) -> List[np.ndarray]:
    """
    Takes a binary image and finds all components (areas with value 1)
    that don't touch the image border.
    """
    height, width = mask.shape
    ret, labels = cv2.connectedComponents(mask)

    indices_to_remove = set()
    for x in range(width):
        indices_to_remove.add(labels[0, x])
        indices_to_remove.add(labels[height - 1, x])
    for y in range(height):
        indices_to_remove.add(labels[y, 0])
        indices_to_remove.add(labels[y, width - 1])
    indices = set(range(1, ret)) - indices_to_remove

    out_masks: List[np.ndarray] = []
    for i in indices:
        out_masks.append(labels == i)
    return out_masks


def get_center_of_component(mask: np.ndarray) -> Tuple[int, int]:
    m = cv2.moments(mask.astype(np.uint8))
    if m["m00"] == 0:
        import matplotlib.pyplot as plt
        plt.imshow(mask)
        plt.show()
    x = int(m["m10"] / m["m00"])
    y = int(m["m01"] / m["m00"])
    return x, y


def point_distance_squared(ax: int, ay: int, bx: int, by: int) -> int:
    """Returns distance between two points squared"""
    return (ax - bx) ** 2 + (ay - by) ** 2


def sort_components_by_proximity_to_point(
        components: List[np.ndarray], x: int, y: int
) -> List[np.ndarray]:
    with_distances = [
        {
            "component": c,
            "distanceSqr": point_distance_squared(
                *get_center_of_component(c),
                x, y
            )
        }
        for c in components
    ]
    with_distances.sort(key=lambda x: x["distanceSqr"])
    return [x["component"] for x in with_distances]


################################################
# Code that actually extracts required symbols #
################################################


def get_full_noteheads(page: MppPage) -> List[Glyph]:
    return _crop_objects_to_single_sprite_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname == "notehead-full"
            and not page.has_outlink_to(o, "ledger_line")
        ],
        page=page,
        label=SmuflLabels.noteheadBlack.value
    )


def get_empty_noteheads(page: MppPage) -> List[Glyph]:
    return _crop_objects_to_single_sprite_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname == "notehead-empty"
            and not page.has_outlink_to(o, "ledger_line")
        ],
        page=page,
        label=SmuflLabels.noteheadWhole.value
    )


def get_whole_rests(page: MppPage) -> List[Glyph]:
    glyphs = _get_symbols_centered_on_line(
        page,
        clsname="whole_rest",
        label=SmuflLabels.restWhole.value,
        line_from_top=1
    )
    # NOTE: these checks were in the old version, but they make situation
    # worse, because some rests are so big that they should be over the staff,
    # not touching the staff. So it's better to keep the original center.
    # for glyph in glyphs:
    #     origin = glyph.sprites[0].bitmap_origin
    #     if origin.y < -0.5 or origin.y > 0.5:
    #         glyph.sprites[0].bitmap_origin = Point(origin.x, 0.0)
    return glyphs


def get_half_rests(page: MppPage) -> List[Glyph]:
    glyphs = _get_symbols_centered_on_line(
        page,
        clsname="half_rest",
        label=SmuflLabels.restHalf.value,
        line_from_top=2
    )
    # NOTE: these checks were in the old version, but they make situation
    # worse, because some rests are so big that they should be over the staff,
    # not touching the staff. So it's better to keep the original center.
    # for glyph in glyphs:
    #     origin = glyph.sprites[0].bitmap_origin
    #     if origin.y < -1.5 or origin.y > 0.5:
    #         glyph.sprites[0].bitmap_origin = Point(origin.x, 1.0)
    return glyphs


def get_quarter_rests(page: MppPage) -> List[Glyph]:
    return _get_symbols_centered_on_line(
        page,
        clsname="quarter_rest",
        label=SmuflLabels.restQuarter.value,
        line_from_top=2,
        when_center_outside_recenter=True
    )


def get_eighth_rests(page: MppPage) -> List[Glyph]:
    return _get_symbols_centered_on_line(
        page,
        clsname="8th_rest",
        label=SmuflLabels.rest8th.value,
        line_from_top=2,
        when_center_outside_recenter=True
    )


def get_sixteenth_rests(page: MppPage) -> List[Glyph]:
    return _get_symbols_centered_on_line(
        page,
        clsname="16th_rest",
        label=SmuflLabels.rest16th.value,
        line_from_top=2,
        when_center_outside_recenter=True
    )


def get_normal_barlines(page: MppPage) -> List[Glyph]:
    _EXCLUDE = set([
        # this is a double barline, accidentally annotated as simple
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-32_N-09_D-ideal___70"
    ])
    return _crop_objects_to_single_sprite_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname in ["thin_barline"]
            and o.height < TALL_BARLINE_THRESHOLD_PX
            and o.uid not in _EXCLUDE
        ],
        page=page,
        label=SmuflLabels.barlineSingle.value
    )


def get_g_clefs(page: MppPage) -> List[Glyph]:
    return _get_symbols_centered_on_line(
        page,
        clsname="g-clef",
        label=SmuflLabels.gClef.value,
        line_from_top=3
    )


def get_f_clefs(page: MppPage) -> List[Glyph]:
    return _get_symbols_centered_on_line(
        page,
        clsname="f-clef",
        label=SmuflLabels.fClef.value,
        line_from_top=1
    )


def get_c_clefs(page: MppPage) -> List[Glyph]:
    return _crop_objects_to_single_sprite_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname == "c-clef"
        ],
        page=page,
        label=SmuflLabels.cClef.value
    )


def get_stems(page: MppPage) -> List[LineGlyph]:
    return _crop_objects_to_line_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname in ["stem"]
        ],
        page=page,
        label=SmuflLabels.stem.value,
        horizontal_line=False, # vertical line
        in_increasing_direction=False # pointing upwards
    )


def get_beams(page: MppPage) -> List[LineGlyph]:
    return _crop_objects_to_line_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname in ["beam"]
            and o.width > BEAM_HOOK_MAX_WIDTH_PX
        ],
        page=page,
        label=SmashcimaLabels.beam.value,
        horizontal_line=True, # horizontal line
        in_increasing_direction=True # pointing to the right
    )


def get_beam_hooks(page: MppPage) -> List[LineGlyph]:
    return _crop_objects_to_line_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname in ["beam"]
            and o.width <= BEAM_HOOK_MAX_WIDTH_PX
        ],
        page=page,
        label=SmashcimaLabels.beamHook.value,
        horizontal_line=True, # horizontal line
        in_increasing_direction=True # pointing to the right
    )


def get_ledger_lines(page: MppPage) -> List[LineGlyph]:
    return _crop_objects_to_line_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname in ["ledger_line"]
        ],
        page=page,
        label=SmashcimaLabels.ledgerLine.value,
        horizontal_line=True, # horizontal line
        in_increasing_direction=True # pointing to the right
    )


def get_flags(page: MppPage) -> Tuple[List[Glyph], List[Glyph]]:
    # NOTE: There are no 32nd flags in the whole MUSCIMA++ dataset.
    # Only 8th and 16th.

    # === extract stem-flag-flag triplets ===

    # stem-id --> (flag-8th, flag-16th)
    extracted_triplets: Dict[
        int,
        Tuple[Optional[CropObject], Optional[CropObject]]
    ] = dict()

    for o in page.crop_objects:
        if o.clsname not in ["8th_flag", "16th_flag"]:
            continue

        # get the stem through a notehead (any, of it's a chord)
        # (ignore empty noteheads, since there will not be flags)
        if not page.has_inlink_from(o, "notehead-full"):
            continue # ignore weird cases
        notehead = page.get_inlink_from(o, "notehead-full")

        # there will be only one stem, because there is no piece in MPP
        # that would have double-stemmed flag chords
        if not page.has_outlink_to(notehead, "stem"):
            continue # ignore weird cases
        stem = page.get_outlink_to(notehead, "stem")

        # store the flag in the triplets
        extracted_triplets.setdefault(stem.objid, (None, None))
        flag8th, flag16th = extracted_triplets[stem.objid]
        
        if o.clsname == "8th_flag":
            flag8th = o
        elif o.clsname == "16th_flag":
            flag16th = o
        
        extracted_triplets[stem.objid] = (flag8th, flag16th)
    
    # === convert stem-flag-flag triplets to composite flags ===

    glyphs_8th_flag: List[Glyph] = []
    glyphs_16th_flag: List[Glyph] = []
    
    for stem_objid, (flag8th, flag16th) in extracted_triplets.items():
        if flag8th is None:
            continue # due to some error, 8th flag was not present 

        _LABEL_LOOKUP: Dict[Tuple[bool, bool], str] = {
            # (16th?, upward?) -> smufl class
            (False, False): SmuflLabels.flag8thDown.value,
            (False, True): SmuflLabels.flag8thUp.value,
            (True, False): SmuflLabels.flag16thDown.value,
            (True, True): SmuflLabels.flag16thUp.value,
        }
        _ISOLATED_LABEL_LOOKUP: Dict[Tuple[bool, bool], str] = {
            # (16th?, upward?) -> smufl class
            (False, False): SmashcimaLabels.isolatedFlag8thDown.value,
            (False, True): SmashcimaLabels.isolatedFlag8thUp.value,
            (True, False): SmashcimaLabels.isolatedFlag16thDown.value,
            (True, True): SmashcimaLabels.isolatedFlag16thUp.value,
        }

        stem = page.get(stem_objid)

        # determine flag orientation
        stem_center_y = (stem.top + stem.bottom) // 2
        flag_center_y = (flag8th.top + flag8th.bottom) // 2
        is_upward_pointing = stem_center_y >= flag_center_y
        is_16th_flag = flag16th is not None

        # get the flag origin in global pixel coordinates
        stem_points = page.point_cloud.get_points(stem)
        if is_upward_pointing:
            stem_tip_point = stem_points[1]
        else:
            stem_tip_point = stem_points[0]
        global_flag_origin = stem_tip_point.vector + \
            Vector2(stem.left, stem.top)

        # create the composed glyph
        flag_subglyphs: List[Glyph] = []

        def _build_sub_glyph(flag: CropObject, label: str):
            local_flag_origin = (
                global_flag_origin - Vector2(flag.left, flag.top)
            )
            relative_flag_origin = Point(
                local_flag_origin.x / flag.width,
                local_flag_origin.y / flag.height
            )
            space = AffineSpace()
            sprite = Sprite(
                space=space,
                bitmap=_mpp_mask_to_sprite_bitmap(flag.mask),
                bitmap_origin=relative_flag_origin,
                dpi=MUSCIMA_PP_DPI
            )
            sub_glyph = Glyph(
                space=space,
                region=Glyph.build_region_from_sprites_alpha_channel(
                    label=label,
                    sprites=[sprite]
                ),
                sprites=[sprite]
            )
            MppGlyphMetadata.stamp_glyph(sub_glyph, page, int(flag.objid))
            flag_subglyphs.append(sub_glyph)
        
        # add sub-glyphs
        _build_sub_glyph(
            flag8th,
            _ISOLATED_LABEL_LOOKUP[False, is_upward_pointing]
        )
        if is_16th_flag:
            _build_sub_glyph(
                flag16th,
                _ISOLATED_LABEL_LOOKUP[True, is_upward_pointing]
            )

        # finalize the composed glyph
        flag_glyph = ComposedGlyph.build(
            label=_LABEL_LOOKUP[is_16th_flag, is_upward_pointing],
            sub_glyphs=flag_subglyphs
        )
        MppGlyphMetadata.stamp_glyph(
            flag_glyph,
            page,
            int(flag8th.objid) # the composite gets the same ID as the 8th flag
        )
        
        # collect up and split by type
        if flag16th is None:
            glyphs_8th_flag.append(flag_glyph)
        else:
            glyphs_16th_flag.append(flag_glyph)

    return glyphs_8th_flag, glyphs_16th_flag


def get_duration_dots(page: MppPage) -> List[Glyph]:
    return _crop_objects_to_single_sprite_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname == "duration-dot"
        ],
        page=page,
        label=SmuflLabels.augmentationDot.value
    )


def get_staccato_dots(page: MppPage) -> List[Glyph]:
    return _crop_objects_to_single_sprite_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname == "staccato-dot"
        ],
        page=page,
        label=SmuflLabels.articStaccatoBelow.value
    )


def get_accidentals(page: MppPage) -> List[Glyph]:
    _LABEL_LOOKUP: Dict[str, str] = {
        "sharp": SmuflLabels.accidentalSharp.value,
        "flat": SmuflLabels.accidentalFlat.value,
        "natural": SmuflLabels.accidentalNatural.value,
        "double_sharp": SmuflLabels.accidentalDoubleSharp.value,

        # NOTE: there are no double-flats in MUSCIMA++
        "double_flat": SmuflLabels.accidentalDoubleFlat.value
    }

    crop_objects = [
        o for o in page.crop_objects
        if o.clsname in list(_LABEL_LOOKUP.keys())
    ]

    glyphs: List[Glyph] = []
    
    for o in crop_objects:
        # obtain sprite from the center of the inner component
        # and handle open accidentals by repeatedly dilating
        # (dilate vertically twice as much)
        object_center_x, object_center_y = get_center_of_component(o.mask)
        mask = (o.mask * 255).astype(dtype=np.uint8)
        space = AffineSpace()
        sprite: Optional[Sprite] = None
        for i in range(5):
            components = get_connected_components_not_touching_image_border(
                1 - (mask // 255)
            )
            components = sort_components_by_proximity_to_point(
                components,
                object_center_x,
                object_center_y
            )

            if len(components) == 0:
                kernel = np.ones((5, 3), np.uint8)
                mask = cv2.dilate(mask, kernel, iterations=1)
                continue

            component_center_x, component_center_y = get_center_of_component(
                components[0]
            )
            sprite = Sprite(
                space=space,
                bitmap=_mpp_mask_to_sprite_bitmap(o.mask),
                bitmap_origin=Point(
                    component_center_x / o.width,
                    component_center_y / o.height
                ),
                dpi=MUSCIMA_PP_DPI
            )
            break

        # if it didn't succeed, try pulling the center out by the attached note
        if sprite is None and len(o.inlinks) == 1:
            link = page.get(o.inlinks[0])
            if "notehead" in link.clsname:
                sprite = Sprite(
                    space=space,
                    bitmap=_mpp_mask_to_sprite_bitmap(o.mask),
                    bitmap_origin=Point(
                        0.5,
                        (((link.top + link.bottom) / 2) - o.top) / o.height
                    ),
                    dpi=MUSCIMA_PP_DPI
                )

        # still nothing, so resort to the crudest method possible
        if sprite is None:
            if o.clsname in ["flat", "double_flat"]:
                sprite = Sprite(
                    space=space,
                    bitmap=_mpp_mask_to_sprite_bitmap(o.mask),
                    bitmap_origin=Point(0.5, 0.75),
                    dpi=MUSCIMA_PP_DPI
                )
            else:
                sprite = Sprite(
                    space=space,
                    bitmap=_mpp_mask_to_sprite_bitmap(o.mask),
                    bitmap_origin=Point(0.5, 0.5),
                    dpi=MUSCIMA_PP_DPI
                )
        
        glyph = Glyph(
            space=space,
            region=Glyph.build_region_from_sprites_alpha_channel(
                label=_LABEL_LOOKUP[o.clsname],
                sprites=[sprite]
            ),
            sprites=[sprite]
        )
        MppGlyphMetadata.stamp_glyph(glyph, page, o.objid)
        glyphs.append(glyph)

    return glyphs


def get_brackets_and_braces(page: MppPage) -> List[LineGlyph]:
    brackets = _crop_objects_to_line_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname == "multi-staff_bracket"
        ],
        page=page,
        label=SmuflLabels.bracket.value,
        horizontal_line=False, # vertical line
        in_increasing_direction=True # drawn top-to-bottom
    )
    braces = _crop_objects_to_line_glyphs(
        crop_objects=[
            o for o in page.crop_objects
            if o.clsname == "multi-staff_brace"
        ],
        page=page,
        label=SmuflLabels.brace.value,
        horizontal_line=False, # vertical line
        in_increasing_direction=True # drawn top-to-bottom
    )
    glyphs = brackets + braces

    # mis-annotated classes
    _MISTAKES = set([
        # braces that are falsely labeled as brackets
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-08_N-10_D-ideal___720",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-08_N-10_D-ideal___721",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-08_N-10_D-ideal___722",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-21_N-05_D-ideal___579",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-24_N-18_D-ideal___282",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-24_N-18_D-ideal___283",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-24_N-18_D-ideal___284",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-41_N-03_D-ideal___458",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-42_N-05_D-ideal___576",

        # brackets that are falsely labeled as braces
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-04_N-20_D-ideal___693",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-04_N-20_D-ideal___694",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-21_N-05_D-ideal___580",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-21_N-05_D-ideal___581",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-39_N-20_D-ideal___712",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-39_N-20_D-ideal___713",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-41_N-03_D-ideal___459",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-41_N-03_D-ideal___466",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-42_N-05_D-ideal___577",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-42_N-05_D-ideal___578",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-46_N-20_D-ideal___604",
        "MUSCIMA-pp_1.0___CVC-MUSCIMA_W-46_N-20_D-ideal___714",
    ])
    for g in glyphs:
        meta = MppGlyphMetadata.of_glyph(g)
        if meta.mpp_crop_object_uid in _MISTAKES:
            if g.label == SmuflLabels.brace.value:
                g.label = SmuflLabels.bracket.value
            elif g.label == SmuflLabels.bracket.value:
                g.label = SmuflLabels.brace.value

    return glyphs


def get_time_marks(page: MppPage) -> List[Glyph]:
    _GLYPH_CLASS_LOOKUP: Dict[str, str] = {
        "numeral_0": SmuflLabels.timeSig0.value, # NOT PRESENT
        "numeral_1": SmuflLabels.timeSig1.value, # NOT PRESENT
        "numeral_2": SmuflLabels.timeSig2.value,
        "numeral_3": SmuflLabels.timeSig3.value,
        "numeral_4": SmuflLabels.timeSig4.value,
        "numeral_5": SmuflLabels.timeSig5.value,
        "numeral_6": SmuflLabels.timeSig6.value,
        "numeral_7": SmuflLabels.timeSig7.value,
        "numeral_8": SmuflLabels.timeSig8.value,
        "numeral_9": SmuflLabels.timeSig9.value, # NOT PRESENT
        "whole-time_mark": SmuflLabels.timeSigCommon.value,
        # cut time is not present in MUSCIMA++
    }

    crop_objects = [
        o for o in page.crop_objects
        if o.clsname == "time_signature"
    ]

    glyphs: List[Glyph] = []

    for o in crop_objects:
        for l in o.outlinks:
            outlink = page.get(l)
            if outlink.clsname == "staff":
                continue

            glyphs += _crop_objects_to_single_sprite_glyphs(
                crop_objects=[outlink],
                page=page,
                label=_GLYPH_CLASS_LOOKUP[outlink.clsname]
            )

    return glyphs



from dataclasses import dataclass, field
from ..SceneObject import SceneObject
from typing import List, Optional
from .Measure import Measure
from .Attributes import Attributes


@dataclass
class Part(SceneObject):
    """Represents a single part (an instrument) of a score"""
    
    measures: List[Measure] = field(default_factory=list)
    "A part consists of a list of measures"

    staff_count: int = 1
    "Number of staves for this part (each measure should have all of them)"

    @classmethod
    def of_measure(cls, measure: Measure):
        return cls.of(measure, lambda p: p.measures)

    def append_measure(self, measure: Measure):
        self.measures = [*self.measures, measure]
    
    def compute_event_attributes(self) -> None:
        """Sets attributes for all events based on present attributes changes"""
        attributes: Optional[Attributes] = None

        for measure in self.measures:
            for event in measure.events:
                
                # the very first event
                if attributes is None:
                    if event.attributes_change is None:
                        raise Exception(
                            "The very first event in a part has to " + \
                            "have an attributes change"
                        )
                    attributes = Attributes.from_first_change(
                        staff_count=self.staff_count,
                        change=event.attributes_change
                    )
                    event.attributes = attributes

                # non-first event
                else:
                    if event.attributes_change:
                        attributes = attributes.apply_change(
                            event.attributes_change
                        )
                    event.attributes = attributes

    def validate(self):
        "Runs various consistency validation checks on the entire part"
        self._validate_consistent_staves()
    
    def _validate_consistent_staves(self):
        """All measures must have the same number of staves, ordered properly"""
        for mi, measure in enumerate(self.measures):
            # check count
            assert len(measure.staves) == self.staff_count, \
                f"Measure {mi} has {len(measure.staves)} staves but should " + \
                f"have {self.staff_count}"

            # check order and value
            for si in range(self.staff_count):
                assert measure.staves[si].staff_number == si + 1, \
                    f"Measure {mi} has incorrect staff numbers"



from .StyleDomain import StyleDomain
from typing import TypeVar, Type, Dict


T = TypeVar("T", bound=StyleDomain)


class Styler:
    """Styler is a service responsible for dictating the style a synthetic
    data sample should be synthesized in. It orchestrates a fleet of
    style domains, where each style domain affects one aspect of the style.
    The synthesizers then access these style domains to know which style
    to use during synthesis."""

    def __init__(self):
        self.domains: Dict[Type[T], StyleDomain] = dict()

    def register_domain(self, domain_type: Type[T], domain_instance: T):
        """Register a specific domain instance for a given style domain type.
        This instance will be given to anyone who asks for the resolution
        of that style domain type."""
        assert issubclass(domain_type, StyleDomain)
        assert isinstance(domain_instance, domain_type)
        
        if domain_type in self.domains:
            raise Exception(
                f"The domain {domain_type} has already been registered."
            )
        
        self.domains[domain_type] = domain_instance
    
    def resolve_domain(self, domain_type: Type[T]) -> T:
        """Returns the registered instance of the given style domain"""
        assert issubclass(domain_type, StyleDomain)
        
        if domain_type not in self.domains:
            raise Exception(
                f"The domain {domain_type} has not been registered."
            )
        
        return self.domains[domain_type]

    def pick_style(self):
        """Go over all registered style domains and let each of them
        pick a style."""
        for domain in self.domains.values():
            domain.pick_style()



from typing import List, Tuple

import numpy as np

from smashcima.geometry import (Point, Rectangle, Transform, Vector2, mm_to_px,
                                px_to_mm)

from .AffineSpace import AffineSpace
from .AffineSpaceVisitor import AffineSpaceVisitor
from .SceneObject import SceneObject


class Sprite(SceneObject):
    """A bitmap image embedded inside the affine space hierarchy
    
    The bitmap is an OpenCV numpy array with BGRA channels and uint8 type.
    
    The sprite lives its parent affine space. The sprite's transform determines,
    where in the affine space lands the bitmap origin point (and possibly
    its orientation and scale). The bitmap origin then specifies, where in the
    bitmap (0.0 to 1.0) is the origin located and the DPI specifies the size
    of a pixel, which in turn defines the physical size of the sprite.

    Another words, a Sprite is a bitmap, with positioning and scaling
    information provided. Scaling is determined by DPI and the origin and
    transform determine the positioning. (Transforms may modify scale in theory,
    but it's a convention that all affine space transforms preserve scale)
    """
    
    def __init__(
        self,
        space: AffineSpace,
        bitmap: np.ndarray,
        bitmap_origin: Point = Point(0.5, 0.5),
        dpi: float = 300,
        transform: Transform = Transform.identity()
    ):
        super().__init__()

        self.space = space
        """The affine space in which the sprite is embedded"""

        self.transform = transform
        """Transform that maps the origin space into the parent space"""

        assert len(bitmap.shape) == 3 # [H, W, C]
        assert bitmap.shape[2] == 4 # BGRA
        assert bitmap.dtype == np.uint8
        self.bitmap = bitmap
        """The numpy opencv BGRA bitmap for the sprite"""

        self.bitmap_origin = bitmap_origin
        """Origin point of the sprite in the normalized pixel space (0.0 - 1.0),
        that is, where does the bitmap overlap with the sprite's transform
        origin"""

        self.dpi = float(dpi)
        """DPI of the bitmap, used for conversion between px and millimeters.
        Together with bitmap resolution determines the sprite's physical size"""
    
    @property
    def pixel_width(self) -> int:
        """Width of the sprite in pixels"""
        return self.bitmap.shape[1]
    
    @property
    def pixel_height(self) -> int:
        """Height of the sprite in pixels"""
        return self.bitmap.shape[0]
    
    @property
    def physical_width(self) -> float:
        """Width of the sprite in millimeters"""
        return px_to_mm(self.pixel_width, dpi=self.dpi)
    
    @property
    def physical_height(self) -> float:
        """Height of the sprite in millimiters"""
        return px_to_mm(self.pixel_height, dpi=self.dpi)
    
    @property
    def pixels_bbox(self) -> Rectangle:
        """Sprite bounding box in its pixel space, that is:
        (0, 0, px_width, px_height)"""
        return Rectangle(0, 0, self.pixel_width, self.pixel_height)
    
    @classmethod
    def many_of_space(cls, space: AffineSpace):
        return cls.many_of(space, lambda s: s.space)
    
    def get_pixels_to_origin_space_transform(self) -> Transform:
        """Returns a transform that converts from local pixel space
        to sprite's origin space (excluding the sprite transform property)"""
        return (
            Transform.translate(Vector2(
                -self.bitmap_origin.x * self.pixel_width,
                -self.bitmap_origin.y * self.pixel_height
            ))
            .then(Transform.scale(px_to_mm(1, dpi=self.dpi)))
        )
    
    def get_pixels_to_parent_space_transform(self) -> Transform:
        """Returns a transform that converts from local pixel space
        to sprite's parent affine space coordinate"""
        return self.get_pixels_to_origin_space_transform().then(self.transform)
    
    def detach(self):
        """Detaches the sprite from the scene hierarchy"""
        self.space = None

    @staticmethod
    def debug_box(
        space: AffineSpace,
        rectangle: Rectangle,
        fill_color = (0, 0, 255, 64), # BGRA
        border_color = (0, 0, 255, 255), # BGRA
        border_width: float = 1,
        dpi: float = 300
    ) -> "Sprite":
        """Creates a rectangular box image sprite with the desired properties"""
        pixel_width = int(mm_to_px(rectangle.width, dpi=dpi))
        pixel_height = int(mm_to_px(rectangle.height, dpi=dpi))
        border_width_px = int(round(mm_to_px(border_width, dpi=dpi)))

        # border is at least 1 px if width is non-zero
        if border_width > 0 and border_width_px == 0:
            border_width_px = 1

        bitmap = np.zeros(
            shape=(pixel_height, pixel_width, 4),
            dtype=np.uint8
        )

        # paint fill
        bitmap[:,:] = fill_color

        # paint border
        if border_width_px > 0:
            bitmap[:border_width_px, :] = border_color
            bitmap[-border_width_px:, :] = border_color
            bitmap[:, :border_width_px] = border_color
            bitmap[:, -border_width_px:] = border_color

        # create the sprite instance
        sprite = Sprite(
            space=space,
            bitmap=bitmap,
            bitmap_origin=Point(0.5, 0.5),
            dpi=dpi
        )
        sprite.transform = Transform.translate(rectangle.center.vector)
        return sprite

    @staticmethod
    def rectangle(
        space: AffineSpace,
        rectangle: Rectangle,
        fill_color = (0, 0, 255, 64), # BGRA
        dpi: float = 300
    ) -> "Sprite":
        """Creates a rectangle filled with the given color at a given DPI"""
        pixel_width = int(mm_to_px(rectangle.width, dpi=dpi))
        pixel_height = int(mm_to_px(rectangle.height, dpi=dpi))
        
        bitmap = np.zeros(
            shape=(pixel_height, pixel_width, 4),
            dtype=np.uint8
        )

        # paint fill
        bitmap[:,:] = fill_color

        # create the sprite instance
        sprite = Sprite(
            space=space,
            bitmap=bitmap,
            bitmap_origin=Point(0.5, 0.5),
            dpi=dpi
        )
        sprite.transform = Transform.translate(rectangle.center.vector)
        return sprite

    @staticmethod
    def traverse_sprites(
        root_space: AffineSpace,
        include_pixels_transform=True,
        include_sprite_transform=True,
        include_root_space_transform=False
    ) -> List[Tuple["Sprite", Transform]]:
        """Returns sprites with composed transforms in a givem space.

        The visiting algorithm starts in the given space (considered the root
        space) and then iterates sprites in the rendering order and returns
        them together with their corresponding transforms that map from the
        pixel space back to the root space.

        :param root_space: The root affine space, where the traversal starts and
            where the returnd transforms map to.
        :param include_pixels_transform: Set to false if the transforms should
            map from the sprite's origin space, not from the pixel space.
        :param include_sprite_transform: Set to false if the transforms should
            map from the sprite's parent affine space, not from the sprite's
            origin space. When this flag is false, the pixels transform must
            also be false.
        :param include_root_space_transform: Set to true if you also want to
            includethe transform from the the root space, to the root's parent
            space.
        """
        visitor = SpriteVisitor(
            space=root_space,
            transform_to_root_space=(
                root_space.transform if include_root_space_transform
                else Transform.identity()
            ),
            include_pixels_transform=include_pixels_transform,
            include_sprite_transform=include_sprite_transform,
        )
        visitor.run()
        return visitor.pairs


class SpriteVisitor(AffineSpaceVisitor):
    """Returns a flat list of sprites and transforms to root space.
    
    Use the `Sprite.traverse_sprites()` function instead of this class.
    """

    def __init__(
        self,
        space: AffineSpace,
        transform_to_root_space: Transform,
        include_pixels_transform=True,
        include_sprite_transform=True,
    ):
        super().__init__(space)

        if include_pixels_transform and not include_sprite_transform:
            raise ValueError(
                "You cannot include pixels transform, but not the sprite " +
                "transform. The resulting transform would be nonsensical."
            )
        
        self.transform_to_root_space = transform_to_root_space
        self.include_pixels_transform = include_pixels_transform
        self.include_sprite_transform = include_sprite_transform

        self.pairs: List[Tuple[Sprite, Transform]] = []

    def create_sub_visitor(self, sub_space: AffineSpace) -> "SpriteVisitor":
        return SpriteVisitor(
            space=sub_space,
            transform_to_root_space=sub_space.transform.then(
                self.transform_to_root_space
            ),
            include_pixels_transform=self.include_pixels_transform,
            include_sprite_transform=self.include_sprite_transform,
        )

    def accept_sub_visitor(self, sub_visitor: "SpriteVisitor"):
        self.pairs += sub_visitor.pairs

    def visit_scene_object(self, obj: SceneObject):
        if isinstance(obj, Sprite):
            sprite = obj

            transform = Transform.identity()
            if self.include_pixels_transform:
                transform = transform.then(
                    sprite.get_pixels_to_origin_space_transform()
                )
            if self.include_sprite_transform:
                transform = transform.then(sprite.transform)
            
            full_transform = transform.then(self.transform_to_root_space)

            self.pairs.append((obj, full_transform))



from .glyph import *
from .layout import *
from .page import *
from .style import *

from .GlyphSynthesizer import GlyphSynthesizer
from .LineSynthesizer import LineSynthesizer

from smashcima.synthesis import glyph
from smashcima.synthesis import layout
from smashcima.synthesis import page
from smashcima.synthesis import style



from typing import TypeVar, Type, Optional, List, Dict
from .AffineSpace import AffineSpace
from .SceneObject import SceneObject


T = TypeVar("T", bound=SceneObject)


class Scene:
    """A scene is a collection of scene objects"""

    def __init__(self, root_space: AffineSpace):
        assert root_space.parent_space is None, \
            "Root space of a scene must have no parent"

        self.root_space = root_space
        "The global space of the scene (parent of all other spaces)"

        self.objects: Dict[int, SceneObject] = {}
        "Tracks all scene objects"

        # add the root space into the scene as a scene object
        self.add(self.root_space)
    
    def has(self, obj: SceneObject) -> bool:
        return id(obj) in self.objects

    def add(
        self,
        obj: SceneObject,
        skip_if_added=True,
        recurse_via_inlinks=True,
        recurse_via_outlinks=True
    ):
        if skip_if_added and self.has(obj):
            return

        self.objects[id(obj)] = obj
        
        # recursion down
        if recurse_via_outlinks:
            for link in obj.outlinks:
                self.add(
                    link.target,
                    skip_if_added=True,
                    recurse_via_inlinks=recurse_via_inlinks,
                    recurse_via_outlinks=recurse_via_outlinks
                )

        # recursion up
        if recurse_via_inlinks:
            for link in obj.inlinks:
                self.add(
                    link.source,
                    skip_if_added=True,
                    recurse_via_inlinks=recurse_via_inlinks,
                    recurse_via_outlinks=recurse_via_outlinks
                )
    
    def add_many(
        self,
        objs: List[SceneObject],
        skip_if_added=True,
        recurse_via_inlinks=True,
        recurse_via_outlinks=True
    ):
        for obj in objs:
            self.add(
                obj,
                skip_if_added=skip_if_added,
                recurse_via_inlinks=recurse_via_inlinks,
                recurse_via_outlinks=recurse_via_outlinks
            )
    
    def add_closure(self):
        """Add all scene objects linked from already added scene objects"""
        for obj in list(self.objects.values()):
            self.add(obj, skip_if_added=False)

    def find(self, obj_type: Type[T]) -> List[T]:
        return [
            obj for obj in self.objects.values()
            if isinstance(obj, obj_type)
        ]



from smashcima.scene.AffineSpace import AffineSpace
from smashcima.scene.Glyph import Glyph
from smashcima.scene.ViewBox import ViewBox
from .BitmapRenderer import BitmapRenderer
import numpy as np
import copy


class DebugGlyphRenderer:
    """Rasterizer meant to render a single glyph with debug metadata overlay"""
    def render(self, glyph: Glyph) -> np.ndarray:
        glyph = copy.deepcopy(glyph) # make a copy and modify that
        
        # add sprites to display debug points
        glyph.place_debug_overlay()

        root_space = AffineSpace()
        glyph.space.parent_space = root_space

        view_box = ViewBox(
            rectangle=glyph.get_bbox_in_space(root_space),
            space=root_space
        )
        dpi = max(s.dpi for s in glyph.sprites)

        renderer = BitmapRenderer(dpi=dpi)
        return renderer.render(view_box)



from .Point import Point
from math import ceil, floor


class Rectangle:
    def __init__(self, x: float, y: float, width: float, height: float):
        assert width >= 0 and height >= 0, "Rectangle cannot have negative size"
        self.x = float(x)
        self.y = float(y)
        self.width = float(width)
        self.height = float(height)
    
    @property
    def left(self) -> float:
        return self.x

    @property
    def top(self) -> float:
        return self.y
    
    @property
    def right(self) -> float:
        return self.x + self.width
    
    @property
    def bottom(self) -> float:
        return self.y + self.height
    
    def __repr__(self):
        return f"Rectangle({self.x}, {self.y}, {self.width}, {self.height})"
    
    @property
    def center(self) -> Point:
        """The center point"""
        return Point(
            self.x + self.width / 2,
            self.y + self.height / 2
        )
    
    @property
    def top_left_corner(self) -> Point:
        """Top left corner point"""
        return Point(self.left, self.top)
    
    @property
    def top_right_corner(self) -> Point:
        """Top right corner point"""
        return Point(self.right, self.top)
    
    @property
    def bottom_left_corner(self) -> Point:
        """Bottom left corner point"""
        return Point(self.left, self.bottom)
    
    @property
    def bottom_right_corner(self) -> Point:
        """Bottom right corner point"""
        return Point(self.right, self.bottom)
    
    @property
    def has_no_area(self) -> bool:
        """Returns true if either of the two dimensions are zero"""
        return self.width == 0 or self.height == 0

    def snap_grow(self) -> "Rectangle":
        """Returns a copy of the rectangle with its size "ceiled" to the
        integer grid. That is, it grows so that it has integer position and size"""
        top = floor(self.top)
        bottom = ceil(self.bottom)
        left = floor(self.left)
        right = ceil(self.right)
        return Rectangle(
            x=left,
            y=top,
            width=right-left,
            height=bottom-top
        )
    
    def snap_shrink(self) -> "Rectangle":
        """Returns a copy of the rectangle with its size "floored to the
        integer grid. That is, it shrinks so that is has integer position and size"""
        top = ceil(self.top)
        bottom = floor(self.bottom)
        left = ceil(self.left)
        right = floor(self.right)
        return Rectangle(
            x=left,
            y=top,
            width=right-left,
            height=bottom-top
        )
    
    def dilate(self, distange: float) -> "Rectangle":
        """Return a rectangle that has grown on all sides uniformly
        by the given distance"""
        top = self.top - distange
        bottom = self.bottom + distange
        left = self.left - distange
        right = self.right + distange
        return Rectangle(
            x=left,
            y=top,
            width=right-left,
            height=bottom-top
        )
    
    def intersect_with(self, other: "Rectangle") -> "Rectangle":
        """Returns the rectangle of the intersection of two rectangles.
        Returns a 0-sized rectangle if the two rectangles don't overlap."""
        left = max(self.left, other.left)
        right = min(self.right, other.right)
        width = max(0, right - left)

        top = max(self.top, other.top)
        bottom = min(self.bottom, other.bottom)
        height = max(0, bottom - top)

        return Rectangle(
            x=left,
            y=top,
            width=width,
            height=height
        )
    
    def relativize_to(self, viewport: "Rectangle") -> "Rectangle":
        """Returns this rectangle with relative coodrinates to the viewport
        
        Relative coordinates vary between [0, 0] and [1, 1] within the viewport,
        but can get outside the bounds if this rectangle is not completely
        within the viewport. Use intersect_with() for clamping.
        """
        return Rectangle(
            x=(self.x - viewport.x) / viewport.width,
            y=(self.y - viewport.y) / viewport.height,
            width=self.width / viewport.width,
            height=self.height / viewport.height
        )
    
    def absolutize_to(self, viewport: "Rectangle") -> "Rectangle":
        """Return rectangle with absolute coordinates given a viewport

        Relative coordinates vary between [0, 0] and [1, 1] within the viewport,
        if `self` is relativized to these coordinates, use this method to
        convert this rectangle's coordinates back to absolute numbers.
        """
        return Rectangle(
            x=viewport.x + self.x * viewport.width,
            y=viewport.y + self.y * viewport.height,
            width=viewport.width * self.width,
            height=viewport.height * self.height
        )



from dataclasses import dataclass
from typing import List

from .Glyph import Glyph
from .ScenePoint import ScenePoint
from .Sprite import Sprite


@dataclass
class LineGlyph(Glyph):
    """Line glyph is a glyph that represents a line segment and as such is
    defined by two point (start and end) placed in the glyph space."""
    
    start_point: ScenePoint
    """Start point of the line, must be set during construction of the glyph.
    Coordinates are relative to the glyph space."""

    end_point: ScenePoint
    """End point of the line, must be set during construction of the glyph.
    Coordinates are relative to the glyph space."""

    @property
    def line_length(self) -> float:
        """Length of the line glyph"""
        return (
            self.end_point.point.vector - self.start_point.point.vector
        ).magnitude

    def place_debug_overlay(self) -> List[Sprite]:
        overlay = super().place_debug_overlay()

        # [red -> blue] is the [start -> end] points
        overlay.append(
            self.start_point.place_debug_overlay(color=(0, 0, 255, 128))
        )
        overlay.append(
            self.end_point.place_debug_overlay(color=(255, 0, 0, 128))
        )

        return overlay



from dataclasses import dataclass
from typing import Dict
from .Clef import Clef
from .TimeSignature import TimeSignature
from .KeySignature import KeySignature
from .AttributesChange import AttributesChange


@dataclass
class Attributes:
    """A state in all the attribute values valid for a given event.
    It's basicaly a running aggregation of AttributesChange objects."""

    staff_count: int
    "Number of staves (in this part)"

    clefs: Dict[int, Clef]
    """Clefs by staff number (staff number starts from 1 for the top staff).
    There's a value for each staff."""

    keys: Dict[int, KeySignature]
    """Key signatures by staff number (staff number starts from 1 for the top
    staff). There's a value for each staff."""

    time_signature: TimeSignature
    "Time signature change, must be at the start of a measure."

    def __post_init__(self):
        assert self.staff_count >= 1, "There must be a positive number of staves"
        assert set(self.clefs.keys()) == set(range(1, self.staff_count+1)), \
            "There is not one clef exactly for each staff"
        assert set(self.keys.keys()) == set(range(1, self.staff_count+1)), \
            "There is not one key signature exactly for each staff"
    
    def apply_change(self, change: AttributesChange) -> "Attributes":
        """Applies a change and returns the resulting attributes"""
        clefs = dict()
        clefs.update(self.clefs)
        clefs.update(change.clefs)

        keys = dict()
        keys.update(self.keys)
        keys.update(change.keys)

        return Attributes(
            staff_count=self.staff_count,
            clefs=clefs,
            keys=keys,
            time_signature=change.time_signature or self.time_signature
        )
    
    @staticmethod
    def from_first_change(
        staff_count: int,
        change: AttributesChange
    ) -> "Attributes":
        """Creates the attributes object from the very first change object"""
        if change.time_signature is None:
            raise Exception(
                "No time signature defined in the first attributes change."
            )
        
        return Attributes(
            staff_count=staff_count,
            clefs=change.clefs, # verified in post-init
            keys=change.keys, # verified in post-init
            time_signature=change.time_signature
        )



from dataclasses import dataclass, field
from ..SceneObject import SceneObject
from typing import List, Set
from .Part import Part
from .ScoreMeasure import ScoreMeasure
from .StaffSemantic import StaffSemantic
from .Measure import Measure
from .Durable import Durable


@dataclass
class Score(SceneObject):
    """Represents a whole music score (a single MusicXML file)"""
    
    parts: List[Part]
    "List of parts that this score consists of"

    new_system_measure_indices: Set[int] = field(default_factory=set)
    "Indices of measures that should be placed on a new system (line breaks)"

    new_page_measure_indices: Set[int] = field(default_factory=set)
    "Indices of measures that should be placed on a new page (page breaks)"

    @property
    def staff_count(self) -> int:
        """How many staves does a single system have"""
        assert len(self.parts) > 0, "There are no parts in the score"
        return sum(p.staff_count for p in self.parts)
    
    @property
    def measure_count(self) -> int:
        """How many measures does the score have in total"""
        assert len(self.parts) > 0, "There are no parts in the score"
        return len(self.parts[0].measures)
    
    def first_staff_index_of_part(self, part: Part) -> int:
        """Given a part returns the staff index (zero-based) of the first staff
        in that part, relative to the system (the score, all the instruments)"""
        index = 0

        for p in self.parts:
            if p is part:
                return index
            else:
                index += p.staff_count

        raise Exception("The given part is not a part of this score.")
    
    def staff_index_of_durable(self, durable: Durable) -> int:
        """Returns the staff index (zero-based) of the given durable
        with respect to the whole score (all the instruments) (to the system)"""
        staff = StaffSemantic.of_durable(durable)
        measure = Measure.of_staff(staff)
        part = Part.of_measure(measure)

        assert part in self.parts, "Given part must be in this score"
        
        return self.first_staff_index_of_part(part) + staff.staff_index
    
    def get_score_measure(self, measure_index: int) -> ScoreMeasure:
        """
        Returns a view of the score from an all-parts temporal perspective.
        This view is constructed each time you call this method and is not
        stored in the data model.
        """
        return ScoreMeasure.from_part_measures([
            self.parts[i].measures[measure_index]
            for i in range(len(self.parts))
        ])

    def validate(self):
        "Runs various consistency validation checks on the entire score"
        self._validate_consistent_measure_count()
        
        for p in self.parts:
            p.validate()
    
    def _validate_consistent_measure_count(self):
        """All parts must have the same number of measures"""
        counts = [len(p.measures) for p in self.parts]
        
        if len(counts) == 0:
            return
        
        first_count = counts[0]
        for c in counts:
            if c != first_count:
                raise Exception(
                    "The score's parts have differing number of measures."
                )



import abc
import json
import shutil
from pathlib import Path
from typing import Any, Dict, Type, TypeVar

from .._version import SMASHCIMA_VERSION_STR

T = TypeVar("T", bound="AssetBundle")


class BundleResolver(abc.ABC):
    """Interface representing something that can resolve asset bundles"""
    
    @abc.abstractmethod
    def resolve_bundle(self, bundle_type: Type[T]) -> T:
        """Ensures that a bundle is installed and returns its instance"""
        raise NotImplementedError


BUNDLE_META_FILE = "bundle.json"


class AssetBundle(abc.ABC):
    """
    Base class representing an asset bundle.
    
    Asset is any piece of external data, which can be used during syntehsis.
    This can be a MUSCIMA++ symbol mask, trained generative ML model,
    or some aggregated distribution.

    Asset bundle is a group of these assets that is downloaded, unpacked,
    and used as a single unit. It is a "library of assets".
    """
    def __init__(
        self,
        bundle_directory: Path,
        dependency_resolver: BundleResolver
    ):
        assert isinstance(bundle_directory, Path)
        self.bundle_directory = bundle_directory
        "Path to the directory where the bundle should be installed"

        self.dependency_resolver = dependency_resolver
        "Use this to resolve additional bundle dependencies"
    
    def version(self) -> Any:
        """Returns version of this bundle. Override this method to modify.
        
        Bundle versions should be incremented to trigger bundle re-install
        for all users who have this bundle already installed. This is useful
        whenever the bundle folder structure changes, or the pickled types
        change.

        Although this should primarily be incrementing integers to trigger
        cache busing, you can return anything that can be serialized to JSON
        and stored in the meta file. Therefore if your bundle requires
        more complex versioning, you can use strings or tuples.
        """
        return 1

    @abc.abstractmethod
    def install(self):
        """Downloads and installs the bundle into the bundle directory."""
        raise NotImplementedError
    
    def remove(self):
        """Removes the bundle from the asset repository folder"""
        shutil.rmtree(self.bundle_directory, ignore_errors=True)
        assert not self.bundle_directory.exists()
    
    def write_metadata(self):
        """Writes the metadata file for the bundle"""
        metadata = {
            # more metadata can be added in the future,
            # such as install datetime, install smashcima version, etc.
            "installed": True,
            "smashcima_version": SMASHCIMA_VERSION_STR,
            "version": self.version()
        }
        with open(self.bundle_directory / BUNDLE_META_FILE, "w") as f:
            json.dump(metadata, f)
    
    def needs_installation(self) -> bool:
        """Returns true if the bundle should be (re-)installed before use"""
        # if there's no metadata file, it probably is not even installed,
        # we definitely need to install before use
        if not self.metadata_exists():
            return True
        
        # load bundle metadata
        with open(self.bundle_directory / BUNDLE_META_FILE, "r") as f:
            metadata: Dict[str, Any] = json.load(f)
        
        # if smashcima version changed, we need re-install, because there
        # might be pickles with old type versions from the library
        if metadata.get("smashcima_version") != SMASHCIMA_VERSION_STR:
            return True
        
        # if bundle version changed, we also need to re-install
        if metadata.get("version") != self.version():
            return True

        return False
    
    def metadata_exists(self) -> bool:
        """Returns true if the metadata file exists for the bundle"""
        return (self.bundle_directory / BUNDLE_META_FILE).is_file()



from typing import List
from smashcima.geometry.Contours import Contours
from smashcima.geometry.Polygon import Polygon
from smashcima.scene.LabeledRegion import LabeledRegion
from smashcima.scene.SmashcimaLabels import SmashcimaLabels
from smashcima.scene.visual.StaffVisual import StaffVisual
from smashcima.scene.visual.StaffCoordinateSystem import StaffCoordinateSystem
from smashcima.scene.AffineSpace import AffineSpace
from smashcima.scene.Sprite import Sprite
from smashcima.scene.Glyph import Glyph
from smashcima.scene.ComposedGlyph import ComposedGlyph
from smashcima.geometry.Vector2 import Vector2
from smashcima.geometry.Rectangle import Rectangle
from smashcima.geometry.Transform import Transform
from smashcima.geometry.units import px_to_mm
from .StafflinesSynthesizer import StafflinesSynthesizer


MUSCIMA_LINE_THICKNESS = px_to_mm(1.5, dpi=300)
MUSCIMA_STAFF_SPACE_UNIT = px_to_mm(28.75, dpi=300)


class NaiveStaffCoordinateSystem(StaffCoordinateSystem):
    def __init__(self, staff_space: float):
        self.staff_space = staff_space
    
    def get_transform(
        self,
        pitch_position: float,
        time_position: float
    ) -> Transform:
        return Transform.translate(
            Vector2(
                x=time_position,
                y=self.staff_space * (-pitch_position / 2)
            )
        )


class NaiveStafflinesSynthesizer(StafflinesSynthesizer):
    """Simple stafflines synthesizer that just creates straight stafflines"""
    def __init__(self) -> None:
        self.staff_space_unit: float = MUSCIMA_STAFF_SPACE_UNIT
        self.line_thickness: float = MUSCIMA_LINE_THICKNESS
        self.line_color = (0, 0, 0, 255)

    @property
    def staff_height(self) -> float:
        return self.staff_space_unit * 4
    
    def synthesize_stafflines(
        self,
        page_space: AffineSpace,
        position: Vector2,
        width: float
    ) -> StaffVisual:
        staff_space = AffineSpace(
            parent_space=page_space,
            transform=Transform.translate(position)
        )

        glyph = self.create_staff_glyph(staff_space, width)

        return StaffVisual(
            width=width,
            staff_coordinate_system=NaiveStaffCoordinateSystem(
                staff_space=self.staff_space_unit
            ),
            space=staff_space,
            glyph=glyph,
            staff_height=self.staff_height
        )
    
    def create_staff_glyph(
        self,
        staff_space: AffineSpace,
        width: float
    ) -> ComposedGlyph:
        sub_glyphs: List[Glyph] = []

        # create all staffline subglyphs
        for line_offset in range(-2, 3):
            line_space = AffineSpace()
            line_rectangle = Rectangle(
                x=0,
                y=line_offset * self.staff_space_unit - self.line_thickness / 2,
                width=width,
                height=self.line_thickness
            )
            line_sprite = Sprite.rectangle(
                line_space,
                line_rectangle,
                fill_color=self.line_color,
                dpi=300
            )
            sub_glyphs.append(Glyph(
                space=line_space,
                region=LabeledRegion(
                    space=line_space,
                    label=SmashcimaLabels.staffLine.value,
                    contours=Contours([
                        Polygon.from_rectangle(line_rectangle)
                    ])
                ),
                sprites=[line_sprite]
            ))
        
        # build the composed glyph
        composed_glyph = ComposedGlyph.build(
            label=SmashcimaLabels.staff.value,
            sub_glyphs=sub_glyphs
        )

        # attach the composed glyph under the staff space
        composed_glyph.space.parent_space = staff_space

        return composed_glyph



from ..SceneObject import SceneObject
from ..LineGlyph import LineGlyph
from ..semantic.Chord import Chord
from ..semantic.BeamedGroup import BeamedGroup
from dataclasses import dataclass
from typing import List


@dataclass
class Beam(SceneObject):
    """One beam (one line) in a beamed group (visually)"""

    glyph: LineGlyph
    "The glyph of the beam"

    beamed_group: BeamedGroup
    "The beamed group this beam belongs to"

    chords: List[Chord]
    """List of chords that are tied together by this beam
    (in chronological order)"""

    beam_number: int
    """Which beam is this beam (1 being the 8th notes beam).
    Uses the MusicXML numbering."""

    def detach(self):
        """Unlink the glyph from the scene"""
        self.glyph.detach()
        self.beamed_group = None
        self.chords = []



from ..LineGlyph import LineGlyph
from ..SceneObject import SceneObject
from ..semantic.Chord import Chord
from dataclasses import dataclass
from typing import Optional
from ..ScenePoint import ScenePoint


@dataclass
class Stem(SceneObject):
    """Stem (visual line) belonging to a chord"""
    
    glyph: LineGlyph
    "The glyph of the line"

    chord: Chord
    """The chord containing the notes that this stem is for. Can be None
    only during construction, otherwise must be set."""

    @property
    def base(self) -> ScenePoint:
        """Base of the stem, in glyph space coordinates"""
        return self.glyph.start_point

    @property
    def tip(self) -> ScenePoint:
        """Tip of the stem, in glyph space coordinates"""
        return self.glyph.end_point
    
    def detach(self):
        """Unlink the glyph from the scene"""
        self.glyph.detach()
        self.chord = None

    @classmethod
    def of_chord(cls, chord: Chord):
        return cls.of(chord, lambda s: s.chord)

    @classmethod
    def of_chord_or_none(cls, chord: Optional[Chord]):
        return cls.of_or_none(chord, lambda s: s.chord)



from ..SceneObject import SceneObject
from dataclasses import dataclass, field


@dataclass
class System(SceneObject):
    """
    A system of music is a single "line" of music with all the instruments
    side by side. It's a collection of staves and measures. It's a visual
    object and does not exist in the musical semantics.
    """
    
    first_measure_index: int
    "What is the index of the first measure on the system (relative to score)"

    measure_count: int
    "Number of measures on this system"

    @property
    def last_measure_index(self) -> int:
        """Returns index of the last measure in the system (relative to score)"""
        return self.first_measure_index + self.measure_count



from dataclasses import dataclass
from typing import List

from ..LabeledRegion import LabeledRegion
from ..SceneObject import SceneObject
from ..semantic.ScoreMeasure import ScoreMeasure
from ..SmashcimaLabels import SmashcimaLabels
from .StaffMeasure import StaffMeasure


@dataclass
class SystemMeasure(SceneObject):
    """Region in the score representing a measure of a system (all staves)"""

    score_measure: ScoreMeasure
    """Semantic measure in the score this visual system measure represents"""

    staff_measures: List[StaffMeasure]
    """Staff measures that make up this system measure"""

    region: LabeledRegion
    """The region encapsulating the system measure"""

    def __post_init__(self):
        assert self.region.label == SmashcimaLabels.systemMeasure.value, \
            "System measure region must have proper classification label"



from .Accidental import Accidental
from .AugmentationDot import AugmentationDot
from .BeamCoordinateSystem import BeamCoordinateSystem
from .BeamHook import BeamHook
from .Beam import Beam
from .Flag import Flag
from .LedgerLine import LedgerLine
from .Notehead import Notehead
from .NoteheadSide import NoteheadSide
from .Page import Page
from .RestVisual import RestVisual
from .StaffCoordinateSystem import StaffCoordinateSystem
from .StaffMeasure import StaffMeasure
from .StaffVisual import StaffVisual
from .Stem import Stem
from .System import System
from .SystemMeasure import SystemMeasure



# -----------------------------------------------------------------------------
# import local types
from .AffineSpace import AffineSpace
from .AffineSpaceVisitor import AffineSpaceVisitor
from .ComposedGlyph import ComposedGlyph
from .Glyph import Glyph
from .LabeledRegion import LabeledRegion
from .LineGlyph import LineGlyph
from .Region import Region
from .Scene import Scene
from .SceneObject import SceneObject
from .ScenePoint import ScenePoint
from .SmashcimaLabels import SmashcimaLabels
from .SmuflLabels import SmuflLabels
from .Sprite import Sprite
from .ViewBox import ViewBox

# -----------------------------------------------------------------------------
# import nested types
from .semantic import *
from .visual import *

# -----------------------------------------------------------------------------
# import sub-modules to make them accessible from this module
from smashcima.scene import semantic
from smashcima.scene import visual



from dataclasses import dataclass
from typing import List

import cv2
import numpy as np

from smashcima.geometry import Contours, Point, Polygon, Rectangle

from .AffineSpace import AffineSpace
from .LabeledRegion import LabeledRegion
from .SceneObject import SceneObject
from .ScenePoint import ScenePoint
from .Sprite import Sprite


@dataclass
class Glyph(SceneObject):
    """Set of sprites together with their label and segmentation region.

    A glyph is a visual unit of the notation. It can be detected, segmented,
    classified.

    It is: notehead, stem, flag, ledger line, staffline,
    But it's also: notehead-stem-flag ligature, ledger-notehead-stem ligature
    Since a ligature cannot be easily broken down into its parts,
    it's a glyph of its own.

    The glyph has its own local coordinate space, relative to which all spatial
    information is represented.

    Also, see the definition of a glyph:
    https://en.wikipedia.org/wiki/Glyph
    """

    space: AffineSpace
    """Space in which glyph components live. The origin is usually well
    defined for each glyph label. The space instance is owned by this
    Glyph instance - should be created, moved, and deleted together."""

    region: LabeledRegion
    """The segmentation mask region, defines the label and bounding box"""

    sprites: List[Sprite]
    """Sprites that make up the visual appearance of the glyph"""

    @property
    def label(self) -> str:
        """Label used for object classification, identifies the type of glyph"""
        return self.region.label
    
    @label.setter
    def label(self, value: str) -> None:
        self.region.label = value
    
    def detach(self):
        """Unlink the glyph from the scene"""
        self.space.parent_space = None # the important bit
        self.region = None
        self.sprites = []
    
    @staticmethod
    def build_region_from_sprites_alpha_channel(
        label: str,
        sprites: List[Sprite],
        threshold: float = 0.5
    ) -> LabeledRegion:
        """Constructs a labeled region from a set of sprites by their alpha.

        All the sprites must already be attached to the same affine space
        and this space will be used as the space of the constructed region.

        This method does not assume overlap in sprites. If resulting contours
        overlap, they will stay overlapping in the resulting region and no
        morphological union will be performed.

        :param label: The classification label that will be assigned to
            the region
        :param sprites: Sprites to use to region construction
        :param threshold: Threshold to use for alpha channel binarization,
            float in 0.0 - 1.0 range.
        """
        assert len(sprites) > 0, "You must provide at least one sprite"

        # get the affine space
        space = sprites[0].space
        assert space is not None, "All provided sprites must have space set"
        assert all(s.space is space for s in sprites), \
            "All provided sprites must be in the same affine space"
        
        contour_polygons: List[Polygon] = []

        # for each sprite
        for sprite in sprites:

            # run contour extraction
            mask = sprite.bitmap[:, :, 3] >= int(threshold * 255)
            img = np.zeros(shape=mask.shape, dtype=np.uint8)
            img[mask] = 255
            cv_contours, _ = cv2.findContours(
                img, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE
            )

            # wrap the results in geometry instances
            transform = sprite.get_pixels_to_origin_space_transform()\
                .then(sprite.transform)
            for cv_contour in cv_contours:
                polygon = transform.apply_to(
                    Polygon.from_cv2_contour(cv_contour)
                )
                contour_polygons.append(polygon)
        
        # build the final region instance
        return LabeledRegion(
            space=space,
            contours=Contours(contour_polygons),
            label=label
        )
    
    def get_bbox_in_space(self, target_space: AffineSpace) -> Rectangle:
        """Returns the bounding box rectangle in the target space coordinates
        
        :param target_space: The space to which coordinates of the contours
            should be transformed. Must be an ancestor of this glyph's space.
        """
        return self.region.get_bbox_in_space(target_space)

    def place_debug_overlay(self) -> List[Sprite]:
        """Places sprites that act as debugging overlay for the glyph.
        Override this method to add more overlay for specialized glyphs."""
        overlay: List[Sprite] = []

        # green origin point
        p = ScenePoint(
            point=Point(0, 0),
            space=self.space
        )
        overlay.append(
            p.place_debug_overlay(color=(0, 255, 0, 128))
        )
        p.detach()
        del p

        return overlay



from smashcima.scene.SmashcimaLabels import SmashcimaLabels
from smashcima.scene.semantic.Score import Score
from smashcima.scene.semantic.Event import Event
from smashcima.scene.semantic.StaffSemantic import StaffSemantic
from smashcima.scene.semantic.Chord import Chord
from smashcima.scene.semantic.StemValue import StemValue
from smashcima.scene.semantic.ScoreEvent import ScoreEvent
from smashcima.scene.semantic.Note import Note
from smashcima.scene.visual.StaffVisual import StaffVisual
from smashcima.scene.visual.Notehead import Notehead
from smashcima.scene.visual.NoteheadSide import NoteheadSide
from smashcima.scene.visual.LedgerLine import LedgerLine
from smashcima.scene.SmuflLabels import SmuflLabels
from smashcima.synthesis.GlyphSynthesizer import GlyphSynthesizer
from smashcima.synthesis.LineSynthesizer import LineSynthesizer
from smashcima.geometry.Point import Point
from smashcima.random_between import random_between
from .ColumnBase import ColumnBase
from typing import List, Dict, Optional, Tuple
from dataclasses import dataclass


# Notehead alignment
# ------------------
# There are 3 columns of noteheads, the center one is the default one and
# if a stem is pointing up, it's placed right of it; if down, it's placed
# left. If there is a one-tone step between two notes, then one of the noteheads
# kicks off to one of the side columns. Which direction it's kicked off depends
# on the stem direction, because the note has to kick off on the other side
# of the stem.


@dataclass
class _NoteheadContext:
    notehead: Notehead
    "The notehead"
    
    note: Note
    "A representative note of that notehead (WLOG the first one)"

    kick_asif_stem_up: bool
    "Perform kick off to the right, else to the left"

    kick_off: int = 0
    "Horizontal notehead column position: -1, 0, +1 for the three columns"


class NoteheadsColumn(ColumnBase):
    def __post_init__(self) -> None:
        self.notehead_contexts: List[_NoteheadContext] = []
        self.ledger_lines: List[LedgerLine] = []
    
    def set_line_synthesizer(self, line_synthesizer: LineSynthesizer):
        self.line_synthesizer = line_synthesizer

    def add_notehead(self, notehead: Notehead):
        assert len(notehead.notes) > 0

        self.glyphs.append(notehead.glyph)
        
        note = notehead.notes[0]
        chord = Chord.of_note(note)

        if chord.stem_value == StemValue.up:
            kick_asif_stem_up = True
        elif chord.stem_value == StemValue.down:
            kick_asif_stem_up = False
        elif chord.stem_value == StemValue.none \
                or chord.stem_value is None: # whole notes or unknown
            kick_asif_stem_up = True # assume invisible up-stem
        else:
            raise Exception(f"Unexpected stem value: {chord.stem_value}")

        self.notehead_contexts.append(_NoteheadContext(
            notehead=notehead,
            note=note,
            kick_asif_stem_up=kick_asif_stem_up
        ))
    
    def _position_glyphs(self):
        # noteheads
        for staff in self.staves:
            self.kick_off_noteheads_on_staff(staff)
        self.position_noteheads()
        
        # ledger lines
        self.delete_current_ledger_lines()
        for staff in self.staves:
            self.place_ledger_lines(staff)

    def kick_off_noteheads_on_staff(self, staff: StaffVisual):
        contexts = [
            c for c in self.notehead_contexts
            if c.notehead.staff is staff
        ]
        contexts.sort(key=lambda c: c.note.pitch.get_linear_pitch())
        
        # lets go from bottom up and kick notes off the center line
        last_linear_pitch = -2
        last_was_kicked_off = True
        for ctx in contexts:
            linear_pitch = ctx.note.pitch.get_linear_pitch()

            # reset kick off
            ctx.kick_off = 0
            ctx.notehead.up_stem_attachment_side = NoteheadSide.right
            ctx.notehead.down_stem_attachment_side = NoteheadSide.left

            # if the previous pitch was kicked off,
            # then the current does not need to be
            if last_was_kicked_off:
                last_was_kicked_off = False
                last_linear_pitch = linear_pitch
                continue

            # ok, now if we have enough space, we also don't need to kick off
            if abs(linear_pitch - last_linear_pitch) >= 2:
                last_was_kicked_off = False
                last_linear_pitch = linear_pitch
                continue

            # now we are definitely tight, we need to kick off
            # just we need to figure out the direction
            last_was_kicked_off = True
            last_linear_pitch = linear_pitch

            if ctx.kick_asif_stem_up:
                ctx.kick_off = 1
                ctx.notehead.up_stem_attachment_side = NoteheadSide.left
                ctx.notehead.down_stem_attachment_side = None
            else:
                ctx.kick_off = -1
                ctx.notehead.up_stem_attachment_side = None
                ctx.notehead.down_stem_attachment_side = NoteheadSide.right

            ctx.kick_off = 1 if ctx.kick_asif_stem_up else -1
        
    def position_noteheads(self):
        if len(self.notehead_contexts) == 0:
            return
        
        center_notehead_stack_width = sum(
            ctx.notehead.glyph.get_bbox_in_space(ctx.notehead.staff.space).width
            for ctx in self.notehead_contexts
        ) / len(self.notehead_contexts) # average
        kick_off_distance = center_notehead_stack_width \
            * 1.2 # scale up a little to accomodate the stem

        for ctx in self.notehead_contexts:
            notehead = ctx.notehead
            sl = ctx.notehead.staff

            notehead.glyph.space.transform = sl.staff_coordinate_system.get_transform(
                pitch_position=ctx.notehead.pitch_position,
                time_position=self.time_position \
                    + (ctx.kick_off * kick_off_distance)
            )
    
    def place_ledger_lines(self, staff: StaffVisual):
        contexts = [
            c for c in self.notehead_contexts
            if c.notehead.staff is staff
        ]

        if len(contexts) == 0:
            return
        
        def _ledger_line_temporal_bounds(
            ctx: _NoteheadContext
        ) -> Tuple[float, float]:
            """Gets the time position of the two start and end of a basic
            ledger line for a given (already placed) notehead"""
            bbox = ctx.notehead.glyph.get_bbox_in_space(ctx.notehead.staff.space)
            pos_x = ctx.notehead.staff.staff_coordinate_system.get_transform(
                pitch_position=0, time_position=self.time_position
            ).apply_to(Point(0, 0)).x
            center = self.time_position + (bbox.center.x - pos_x)
            
            # determine the ledger line width
            # TODO: get from some distribution
            width = bbox.width * random_between(1.2, 2.5, self.rng)
            
            start = center - width / 2
            end = center + width / 2
            return start, end
        
        max_abs_pitch_position = max(
            abs(ctx.notehead.pitch_position) for ctx in contexts
        )

        # top-down, then bottom-up
        for sign in [1, -1]:
            affected_noteheads: List[Notehead] = []
            time_position_start = float("inf")
            time_position_end = float("-inf")

            # walk pitches towards staff center
            # (code is written for going top-down, then inverted by the sign)
            STAFF_EDGE_PP = 4
            for pitch_position in range(
                max_abs_pitch_position * sign, # start
                STAFF_EDGE_PP * sign, # stop (exclusive)
                -1 * sign # step
            ):
                # expand affected noteheads by the current noteheads
                for ctx in contexts:
                    if ctx.notehead.pitch_position == pitch_position:
                        affected_noteheads.append(ctx.notehead)
                        start, end = _ledger_line_temporal_bounds(ctx)
                        time_position_start = min(time_position_start, start)
                        time_position_end = max(time_position_end, end)
                
                # create ledger line
                # (if on even position and if we have some affected noteheads)
                if pitch_position % 2 == 0 and len(affected_noteheads) > 0:
                    line = self.synthesize_ledger_line(
                        staff,
                        pitch_position=pitch_position,
                        time_position_start=time_position_start,
                        time_position_end=time_position_end
                    )
                    line.affected_noteheads = [*affected_noteheads] # copy
    
    def delete_current_ledger_lines(self):
        for line in self.ledger_lines:
            self.glyphs.remove(line.glyph)
            line.detach()
        self.ledger_lines = []
    
    def synthesize_ledger_line(
        self,
        staff: StaffVisual,
        pitch_position: int,
        time_position_start: float,
        time_position_end: float
    ) -> LedgerLine:
        start_point = staff.staff_coordinate_system.get_transform(
            pitch_position=pitch_position,
            time_position=time_position_start
        ).apply_to(Point(0, 0))

        end_point = staff.staff_coordinate_system.get_transform(
            pitch_position=pitch_position,
            time_position=time_position_end
        ).apply_to(Point(0, 0))

        glyph = self.line_synthesizer.synthesize_line(
            label=SmashcimaLabels.ledgerLine.value,
            parent_space=staff.space,
            start_point=start_point,
            end_point=end_point
        )
        line = LedgerLine(
            glyph=glyph,
            affected_noteheads=[], # populated later
            affected_rest=None
        )

        self.glyphs.append(glyph)
        self.ledger_lines.append(line)

        return line


def synthesize_noteheads_column(
    column: NoteheadsColumn,
    staves: List[StaffVisual],
    glyph_synthesizer: GlyphSynthesizer,
    line_synthesizer: LineSynthesizer,
    score: Score,
    score_event: ScoreEvent
):
    column.set_line_synthesizer(line_synthesizer)
    
    # notehead merging logic
    # (when two voices share a notehead)
    _all_noteheads: Dict[Tuple[int, int], Notehead] = dict()
    
    def _get_notehead(note: Note, staff_index: int) -> Optional[Notehead]:
        nonlocal _all_noteheads
        linear_pitch = note.pitch.get_linear_pitch()
        return _all_noteheads.get((staff_index, linear_pitch))

    def _store_notehead(note: Note, staff_index: int, notehead: Notehead):
        nonlocal _all_noteheads
        linear_pitch = note.pitch.get_linear_pitch()
        _all_noteheads[(staff_index, linear_pitch)] = notehead

    # go through all the notes and create notehead glyphs
    for event in score_event.events:
        for durable in event.durables:
            if isinstance(durable, Note):
                note: Note = durable
                staff_index = score.staff_index_of_durable(note)
                notehead = _get_notehead(note, staff_index)

                # another note for an existing notehead
                if notehead is not None:
                    notehead.notes = [*notehead.notes, note]
                    continue

                # resolve context
                event = Event.of_durable(note)
                staff_sem = StaffSemantic.of_durable(note)
                clef = event.attributes.clefs[staff_sem.staff_number]
                staff = staves[staff_index]
                
                # new notehead for a note
                glyph = glyph_synthesizer.synthesize_glyph_at(
                    label=SmuflLabels.notehead_from_type_duration(
                        note.type_duration
                    ).value,
                    parent_space=staff.space,
                    point=Point(0, 0) # glyph positioned later
                )
                notehead = Notehead(
                    glyph=glyph,
                    notes = [note],
                    clef=clef,
                    staff=staff,
                    pitch_position=clef.pitch_to_pitch_position(note.pitch),
                )
                _store_notehead(note, staff_index, notehead)
    
    # add noteheads to the column
    for notehead in _all_noteheads.values():
        column.add_notehead(notehead)



import abc
from typing import Set

import cv2
import numpy as np

from smashcima.geometry import Point, Transform, Vector2
from smashcima.scene import AffineSpace, LineGlyph


class LineSynthesizer(abc.ABC):
    """Interface for a line glyph synthesizer"""

    def synthesize_line(
        self,
        label: str,
        parent_space: AffineSpace,
        start_point: Point,
        end_point: Point,
    ) -> LineGlyph:
        """Synthesizes a line with the given classification label and endpoints
        
        :param label: Which kind of line to synthesize (classification label)
        :param parent_space: The glyph will be placed into this space
        :param start_point: Where the line starts in the parent space
        :param end_point: Where the line ends in the parent space
        :returns: The newly synthesized glyph instance, already attached to
            the parent space.
        """
        # check that we are not accidentally passing in an Enum instead of str
        assert type(label) is str, "The requested glyph label is not a string"

        # compute the delta vector
        delta = end_point.vector - start_point.vector

        # create a new glyph
        glyph = self.create_glyph(
            label=label,
            delta=delta
        )

        # check its label
        assert glyph.label == label, f"The user requested class {label} " + \
            f"but got {glyph.label} instead"

        # compute the transform to apply to the created glyph to place
        # it exactly where the user wants it
        transform = LineSynthesizer.line_to_line_transform_preserving_thickness(
            from_origin=glyph.start_point.point,
            from_delta=(
                glyph.end_point.point.vector - glyph.start_point.point.vector
            ),
            to_origin=start_point,
            to_delta=delta
        )
        
        # attach the glyph to the parent space
        glyph.space.parent_space = parent_space
        glyph.space.transform = transform

        return glyph
    
    @staticmethod
    def line_to_line_transform_preserving_thickness(
        from_origin: Point,
        from_delta: Vector2,
        to_origin: Point,
        to_delta: Vector2
    ) -> Transform:
        """Computes transform mapping from one line segment to another.
        
        The segment thickness (the perpendicular direction) is preserved.

        :param from_origin: Starting point of the starting line segment.
        :param from_delta: Direction vector of the starting line segment.
        :param to_origin: Starting point of the target line segment.
        :param to_delta: Direction vector of the target line segment.
        """
        # frame is [origin, X unit, Y unit]
        from_frame = np.array([
            list(from_origin),
            list(from_origin.vector + from_delta),
            list(from_origin.vector + from_delta.normalize().rotate90degCC())
        ], dtype=np.float32)

        to_frame = np.array([
            list(to_origin),
            list(to_origin.vector + to_delta),
            list(to_origin.vector + to_delta.normalize().rotate90degCC())
        ], dtype=np.float32)

        matrix = cv2.getAffineTransform(from_frame, to_frame)

        return Transform(matrix)
    
    def create_glyph(self, label: str, delta: Vector2) -> LineGlyph:
        """Creates a new glyph instance of the specified label.

        The delta vector is used as a recommendation, not a command. It is
        the vector that the user wants to synthesize, but the created glyph
        will have its transform updated so that it matches the user's
        requested position and scale. But you can use this delta vector to
        create a glyph with the desired length / thickness / orientation.

        Override this method when building custom glyph synthesizers.
        The returned glyph should have no parent space.

        :param label: The kind of glyph to create.
        :param delta: Delta vector from the start point to the end point.
        """
        raise NotImplementedError



from enum import Enum
from .semantic import TypeDuration,StemValue, AccidentalValue


class SmuflLabels(str, Enum):
    """Classification labels from the SMuFL specification.
    
    Learn more at: https://www.w3.org/2019/03/smufl13/
    """

    # Brackets and dividers
    # https://www.w3.org/2019/03/smufl13/tables/staff-brackets-and-dividers.html
    brace = "smufl::brace"
    bracket = "smufl::bracket"

    # Staves
    # https://www.w3.org/2019/03/smufl13/tables/staves.html
    # IMPORTANT: The stafflines and ledger lines section in SMuFL is intended
    # only for text-based applications. Use Smashcima label to represent
    # staves and stafflines instead.

    # Barlines
    # https://w3c.github.io/smufl/latest/tables/barlines.html
    barlineSingle = "smufl::barlineSingle"

    # Clefs
    # https://w3c.github.io/smufl/latest/tables/clefs.html
    gClef = "smufl::gClef"
    cClef = "smufl::cClef"
    fClef = "smufl::fClef"
    gClefSmall = "smufl::gClefSmall"
    cClefSmall = "smufl::cClefSmall"
    fClefSmall = "smufl::fClefSmall"

    # Time signatures
    # https://www.w3.org/2019/03/smufl13/tables/time-signatures.html
    timeSig0 = "smufl::timeSig0"
    timeSig1 = "smufl::timeSig1"
    timeSig2 = "smufl::timeSig2"
    timeSig3 = "smufl::timeSig3"
    timeSig4 = "smufl::timeSig4"
    timeSig5 = "smufl::timeSig5"
    timeSig6 = "smufl::timeSig6"
    timeSig7 = "smufl::timeSig7"
    timeSig8 = "smufl::timeSig8"
    timeSig9 = "smufl::timeSig9"
    timeSigCommon = "smufl::timeSigCommon"
    timeSigCutCommon = "smufl::timeSigCutCommon"

    # Noteheads
    # https://w3c.github.io/smufl/latest/tables/noteheads.html
    noteheadDoubleWhole = "smufl::noteheadDoubleWhole"
    noteheadDoubleWholeSquare = "smufl::noteheadDoubleWholeSquare"
    noteheadWhole = "smufl::noteheadWhole"
    noteheadHalf = "smufl::noteheadHalf"
    noteheadBlack = "smufl::noteheadBlack"

    # Individual Notes
    # https://www.w3.org/2019/03/smufl13/tables/individual-notes.html
    # IMPORTANT: This should only be used for ligatures,
    # the default usecase is synthesizing notehead-stem-flag separately.
    noteWhole = "smulf::noteWhole"
    noteHalfUp = "smulf::noteHalfUp"
    noteHalfDown = "smulf::noteHalfDown"
    noteQuarterUp = "smulf::noteQuarterUp"
    noteQuarterDown = "smulf::noteQuarterDown"
    note8thUp = "smulf::note8thUp"
    note8thDown = "smulf::note8thDown"
    note16thUp = "smulf::note16thUp"
    note16thDown = "smulf::note16thDown"
    note32ndUp = "smulf::note32ndUp"
    note32ndDown = "smulf::note32ndDown"
    note64thUp = "smufl::note64thUp"
    note64thDown = "smufl::note64thDown"
    note128thUp = "smufl::note128thUp"
    note128thDown = "smufl::note128thDown"
    note256thUp = "smufl::note256thUp"
    note256thDown = "smufl::note256thDown"
    note512thUp = "smufl::note512thUp"
    note512thDown = "smufl::note512thDown"
    note1024thUp = "smufl::note1024thUp"
    note1024thDown = "smufl::note1024thDown"
    augmentationDot = "smulf::augmentationDot"

    # Stems
    # https://w3c.github.io/smufl/latest/tables/stems.html
    stem = "smufl::stem"

    # Flags
    # https://w3c.github.io/smufl/latest/tables/flags.html
    flag8thUp = "smufl::flag8thUp"
    flag8thDown = "smufl::flag8thDown"
    flag16thUp = "smufl::flag16thUp"
    flag16thDown = "smufl::flag16thDown"
    flag32ndUp = "smufl::flag32ndUp"
    flag32ndDown = "smufl::flag32ndDown"
    flag64thUp = "smufl::flag64thUp"
    flag64thDown = "smufl::flag64thDown"
    flag128thUp = "smufl::flag128thUp"
    flag128thDown = "smufl::flag128thDown"
    flag256thUp = "smufl::flag256thUp"
    flag256thDown = "smufl::flag256thDown"
    flag512thUp = "smufl::flag512thUp"
    flag512thDown = "smufl::flag512thDown"
    flag1024thUp = "smufl::flag1024thUp"
    flag1024thDown = "smufl::flag1024thDown"

    # Accidentals
    # https://www.w3.org/2019/03/smufl13/tables/standard-accidentals-12-edo.html
    accidentalFlat = "smufl::accidentalFlat"
    accidentalNatural = "smufl::accidentalNatural"
    accidentalSharp = "smufl::accidentalSharp"
    accidentalDoubleSharp = "smufl::accidentalDoubleSharp"
    accidentalDoubleFlat = "smufl::accidentalDoubleFlat"
    accidentalTripleSharp = "smufl::accidentalTripleSharp"
    accidentalTripleFlat = "smufl::accidentalTripleFlat"
    accidentalNaturalFlat = "smufl::accidentalNaturalFlat"
    accidentalNaturalSharp = "smufl::accidentalNaturalSharp"
    accidentalSharpSharp = "smufl::accidentalSharpSharp"
    accidentalParensLeft = "smufl::accidentalParensLeft"
    accidentalParensRight = "smufl::accidentalParensRight"
    accidentalBracketLeft = "smufl::accidentalBracketLeft"
    accidentalBracketRight = "smufl::accidentalBracketRight"

    # Articulation
    # https://www.w3.org/2019/03/smufl13/tables/articulation.html
    articAccentAbove = "smufl::articAccentAbove"
    articAccentBelow = "smufl::articAccentBelow"
    articStaccatoAbove = "smufl::articStaccatoAbove"
    articStaccatoBelow = "smufl::articStaccatoBelow"
    articTenutoAbove = "smufl::articTenutoAbove"
    articTenutoBelow = "smufl::articTenutoBelow"
    articStaccatissimoAbove = "smufl::articStaccatissimoAbove"
    articStaccatissimoBelow = "smufl::articStaccatissimoBelow"
    articStaccatissimoWedgeAbove = "smufl::articStaccatissimoWedgeAbove"
    articStaccatissimoWedgeBelow = "smufl::articStaccatissimoWedgeBelow"
    articStaccatissimoStrokeAbove = "smufl::articStaccatissimoStrokeAbove"
    articStaccatissimoStrokeBelow = "smufl::articStaccatissimoStrokeBelow"
    articMarcatoAbove = "smufl::articMarcatoAbove"
    articMarcatoBelow = "smufl::articMarcatoBelow"
    articMarcatoStaccatoAbove = "smufl::articMarcatoStaccatoAbove"
    articMarcatoStaccatoBelow = "smufl::articMarcatoStaccatoBelow"
    articAccentStaccatoAbove = "smufl::articAccentStaccatoAbove"
    articAccentStaccatoBelow = "smufl::articAccentStaccatoBelow"
    articTenutoStaccatoAbove = "smufl::articTenutoStaccatoAbove"
    articTenutoStaccatoBelow = "smufl::articTenutoStaccatoBelow"
    articTenutoAccentAbove = "smufl::articTenutoAccentAbove"
    articTenutoAccentBelow = "smufl::articTenutoAccentBelow"
    articStressAbove = "smufl::articStressAbove"
    articStressBelow = "smufl::articStressBelow"
    articUnstressAbove = "smufl::articUnstressAbove"
    articUnstressBelow = "smufl::articUnstressBelow"
    articLaissezVibrerAbove = "smufl::articLaissezVibrerAbove"
    articLaissezVibrerBelow = "smufl::articLaissezVibrerBelow"
    articMarcatoTenutoAbove = "smufl::articMarcatoTenutoAbove"
    articMarcatoTenutoBelow = "smufl::articMarcatoTenutoBelow"

    # Rests
    # https://www.w3.org/2019/03/smufl13/tables/rests.html
    restMaxima = "smulf::restMaxima"
    restLonga = "smulf::restLonga"
    restDoubleWhole = "smulf::restDoubleWhole"
    restWhole = "smulf::restWhole"
    restHalf = "smulf::restHalf"
    restQuarter = "smulf::restQuarter"
    rest8th = "smulf::rest8th"
    rest16th = "smulf::rest16th"
    rest32nd = "smulf::rest32nd"
    rest64th = "smulf::rest64th"
    rest128th = "smulf::rest128th"
    rest256th = "smulf::rest256th"
    rest512th = "smulf::rest512th"
    rest1024th = "smulf::rest1024th"

    @staticmethod
    def notehead_from_type_duration(duration: TypeDuration) -> "SmuflLabels":
        # https://www.w3.org/2021/06/musicxml40/musicxml-reference/data-types/note-type-value/
        _LOOKUP = {
            TypeDuration.thousand_twenty_fourth: SmuflLabels.noteheadBlack,
            TypeDuration.five_hundred_twelfth: SmuflLabels.noteheadBlack,
            TypeDuration.two_hundred_fifty_sixth: SmuflLabels.noteheadBlack,
            TypeDuration.hundred_twenty_eighth: SmuflLabels.noteheadBlack,
            TypeDuration.sixty_fourth: SmuflLabels.noteheadBlack,
            TypeDuration.thirty_second: SmuflLabels.noteheadBlack,
            TypeDuration.sixteenth: SmuflLabels.noteheadBlack,
            TypeDuration.eighth: SmuflLabels.noteheadBlack,
            TypeDuration.quarter: SmuflLabels.noteheadBlack,
            TypeDuration.half: SmuflLabels.noteheadHalf,
            TypeDuration.whole: SmuflLabels.noteheadWhole,
            TypeDuration.breve: SmuflLabels.noteheadDoubleWhole,
            TypeDuration.long: SmuflLabels.noteheadDoubleWholeSquare,
            TypeDuration.maxima: SmuflLabels.noteheadDoubleWholeSquare,
        }
        notehead = _LOOKUP.get(duration)
        if notehead is None:
            raise Exception(f"Unsupported type duration " + repr(duration))
        return notehead
    
    @staticmethod
    def rest_from_type_duration(duration: TypeDuration) -> "SmuflLabels":
        # https://www.w3.org/2021/06/musicxml40/musicxml-reference/data-types/note-type-value/
        _LOOKUP = {
            TypeDuration.thousand_twenty_fourth: SmuflLabels.rest1024th,
            TypeDuration.five_hundred_twelfth: SmuflLabels.rest512th,
            TypeDuration.two_hundred_fifty_sixth: SmuflLabels.rest256th,
            TypeDuration.hundred_twenty_eighth: SmuflLabels.rest128th,
            TypeDuration.sixty_fourth: SmuflLabels.rest64th,
            TypeDuration.thirty_second: SmuflLabels.rest32nd,
            TypeDuration.sixteenth: SmuflLabels.rest16th,
            TypeDuration.eighth: SmuflLabels.rest8th,
            TypeDuration.quarter: SmuflLabels.restQuarter,
            TypeDuration.half: SmuflLabels.restHalf,
            TypeDuration.whole: SmuflLabels.restWhole,
            TypeDuration.breve: SmuflLabels.restDoubleWhole,
            TypeDuration.long: SmuflLabels.restLonga,
            TypeDuration.maxima: SmuflLabels.restMaxima,
        }
        notehead = _LOOKUP.get(duration)
        if notehead is None:
            raise Exception(f"Unsupported type duration " + repr(duration))
        return notehead
    
    @staticmethod
    def clef_from_clef_sign(clef_sign: str, small=False) -> "SmuflLabels":
        _LOOKUP = {
            ("G", False): SmuflLabels.gClef,
            ("G", True): SmuflLabels.gClefSmall,
            ("F", False): SmuflLabels.fClef,
            ("F", True): SmuflLabels.fClefSmall,
            ("C", False): SmuflLabels.cClef,
            ("C", True): SmuflLabels.cClefSmall,
        }
        key = (clef_sign, small)
        clef = _LOOKUP.get(key)
        if clef is None:
            raise Exception(f"Unsupported clef " + repr(key))
        return clef

    @staticmethod
    def flag_from_type_duration_and_stem_value(
        type_duration: TypeDuration,
        stem_value: StemValue
    ) -> "SmuflLabels":
        _LOOKUP = {
            ("up", "eighth"): SmuflLabels.flag8thUp,
            ("down", "eighth"): SmuflLabels.flag8thDown,
            ("up", "16th"): SmuflLabels.flag16thUp,
            ("down", "16th"): SmuflLabels.flag16thDown,
            ("up", "32nd"): SmuflLabels.flag32ndUp,
            ("down", "32nd"): SmuflLabels.flag32ndDown,
            ("up", "64th"): SmuflLabels.flag64thUp,
            ("down", "64th"): SmuflLabels.flag64thDown,
            ("up", "128th"): SmuflLabels.flag128thUp,
            ("down", "128th"): SmuflLabels.flag128thDown,
            ("up", "256th"): SmuflLabels.flag256thUp,
            ("down", "256th"): SmuflLabels.flag256thDown,
            ("up", "512th"): SmuflLabels.flag512thUp,
            ("down", "512th"): SmuflLabels.flag512thDown,
            ("up", "1024th"): SmuflLabels.flag1024thUp,
            ("down", "1024th"): SmuflLabels.flag1024thDown,
        }
        key = (stem_value.value, type_duration.value)
        flag = _LOOKUP.get(key)
        if flag is None:
            raise Exception(f"Unsupported flag " + repr(key))
        return flag

    @staticmethod
    def accidental_from_accidental_value(
        accidental_value: AccidentalValue
    ) -> "SmuflLabels":
        _LOOKUP = {
            AccidentalValue.natural: SmuflLabels.accidentalNatural,
            AccidentalValue.flat: SmuflLabels.accidentalFlat,
            AccidentalValue.sharp: SmuflLabels.accidentalSharp,
            AccidentalValue.doubleSharp: SmuflLabels.accidentalDoubleSharp,
        }
        label = _LOOKUP.get(accidental_value)
        if label is None:
            raise Exception(f"Unsupported accidental " + repr(accidental_value))
        return label



from .StyleDomain import StyleDomain
from smashcima.assets.AssetRepository import AssetRepository
from smashcima.assets.glyphs.muscima_pp.MuscimaPPGlyphs import MuscimaPPGlyphs
import random
from typing import List


class MuscimaPPStyleDomain(StyleDomain):
    """This style domain represents the set of 50 writers of the MUSCIMA++
    dataset. When sampled, one of these writers is chosen."""

    def __init__(self, assets: AssetRepository, rng: random.Random):
        bundle = assets.resolve_bundle(MuscimaPPGlyphs)
        symbol_repository = bundle.load_symbol_repository()
        
        self.all_writers: List[int] = list(sorted(symbol_repository.all_writers))
        "The domain of all MPP writers (their numbers)"
        
        self.rng = rng
        "The RNG used for randomness"

        assert len(self.all_writers) > 0, "There must be at least one writer"

        # use the first writer as the default
        # (the pick_style will be called before the first sampling)
        self.current_writer: int = self.all_writers[0]
        "The writer number to be used for the currently synthesized sample"

    def pick_style(self):
        self.current_writer = self.rng.choice(self.all_writers)



from dataclasses import dataclass, field
from fractions import Fraction
from ..SceneObject import SceneObject
from .Durable import Durable
from .Event import Event
from .StaffSemantic import StaffSemantic
from .AttributesChange import AttributesChange
from typing import List


@dataclass
class Measure(SceneObject):
    """A measure of music for a single instrument (a single part)"""

    events: List[Event] = field(default_factory=list)
    "Links to all events within this measure"

    staves: List[StaffSemantic] = field(default_factory=list)
    """Links to all staves within this measure"""

    @property
    def first_event(self) -> Event:
        assert self.events[0].fractional_measure_onset == 0, \
            "The first event in the measure does not have 0 onset"
        return self.events[0]

    @staticmethod
    def of_durable(durable: Durable):
        event = Event.of_durable(durable)
        return Measure.of_event(event)

    @classmethod
    def of_event(cls, event: Event):
        return cls.of(event, lambda m: m.events)
    
    @classmethod
    def of_staff(cls, staff: StaffSemantic):
        return cls.of(staff, lambda m: m.staves)
    
    def add_attributes_change(
        self,
        change: AttributesChange,
        onset: Fraction
    ):
        """Adds an attributes change into the measure"""
        event = self.get_or_create_event(onset)
        event.attributes_change = change

    def add_durable(
        self,
        durable: Durable,
        onset: Fraction,
        staff_number: int
    ):
        """Adds a durable into the measure"""
        event = self.get_or_create_event(onset)
        event.durables = [*event.durables, durable]

        staff = self.get_or_create_staff(staff_number)
        staff.durables = [*staff.durables, durable]

    def get_or_create_event(self, onset: Fraction) -> Event:
        """Returns event with the given onset"""
        matches = [e for e in self.events if e.fractional_measure_onset == onset]
        if len(matches) > 1:
            raise Exception("There are multiple events with the same onset!")
        elif len(matches) == 1:
            return matches[0]
        else:
            event = Event(fractional_measure_onset=onset)
            self.events = [*self.events, event]
            self.sort_events_by_onset()
            return event
    
    def sort_events_by_onset(self):
        """Sorts events by onset, ascending"""
        self.events.sort(key=lambda e: e.fractional_measure_onset)
    
    def get_or_create_staff(self, staff_number: int) -> StaffSemantic:
        """Returns staff with the given staff number"""
        matches = [s for s in self.staves if s.staff_number == staff_number]
        if len(matches) > 1:
            raise Exception("There are multiple staves with the same number!")
        elif len(matches) == 1:
            return matches[0]
        else:
            staff = StaffSemantic(staff_number=staff_number)
            self.staves = [*self.staves, staff]
            self.sort_staves_by_number()
            return staff
    
    def sort_staves_by_number(self):
        """Sorts staves by number, ascending"""
        self.staves.sort(key=lambda e: e.staff_number)



from dataclasses import dataclass, field
from ..SceneObject import SceneObject
from fractions import Fraction
from .Durable import Durable
from .Attributes import Attributes
from .AttributesChange import AttributesChange
from typing import List, Optional


@dataclass
class Event(SceneObject):
    """A collection of notes (and other symbols) with the same musical onset"""
    
    fractional_measure_onset: Fraction
    """How many quarter notes from the beginning of the measure this event occus"""

    durables: List[Durable] = field(default_factory=list)
    "Links to all durables with this onset time"

    attributes: Optional[Attributes] = None
    """State of attributes (clefs, keys, time) at this event. Should be null
    only during the parsing of a score. This value is populated when a part
    is parsed by calling Part.compute_event_attributes()"""

    attributes_change: Optional[AttributesChange] = None
    """Change of attributes when entering this event. If None, there is no
    change on this event."""

    @classmethod
    def of_durable(cls, durable: Durable):
        return cls.of(durable, lambda e: e.durables)



from pathlib import Path
from typing import Optional, TextIO, Union

from smashcima.scene.semantic.Score import Score

from .MusicXmlLoader import MusicXmlLoader


def load_score(
    file: Union[Path, str, None] = None,
    data: Union[bytes, str, None] = None,
    format: Optional[str] = None,
    soft_errors_output: Optional[TextIO] = None
) -> Score:
    """Loads a file with music notation into a Smashcima Score object.
    
    :param file: Path to a file with musical content (e.g. './my_file.musicxml')
    :param data: Musical contents as a bytes or string (e.g. MusicXML
        file contents)
    :param format: In what format is the musical content (file suffix,
        including the period, i.e. '.musicxml')
    :param soft_errors_output: Optional output for logging soft errors.
    :returns: The parsed score object
    """

    # get format from the suffix
    if file is not None and format is None:
        format = Path(str(file)).suffix or None

    # there must be some annotation format specified
    if format is None:
        raise Exception(
            "Annotation format cannot be deduced. " + \
            "You need to specify it explicitly."
        )
    
    # either file or data must be provided
    if file is None and data is None:
        raise Exception("Either file path or file data must be provided")

    # select parser by format
    if format in [".mxl"]:
        raise Exception("Compressed MusicXML format is not yet supported")
    elif format in [".musicxml"]:
        mxl_loader = MusicXmlLoader(errout=soft_errors_output)
        if data is not None:
            if type(data) is bytes:
                data = data.decode("utf-8")
            return mxl_loader.load_xml(data)
        else:
            return mxl_loader.load_file(file)
    
    raise Exception(f"Format {format} cannot be loaded")



from dataclasses import dataclass
from typing import List, Optional

from ..Glyph import Glyph
from ..SceneObject import SceneObject
from ..semantic import Clef, Note
from .NoteheadSide import NoteheadSide
from .StaffVisual import StaffVisual


@dataclass
class Notehead(SceneObject):
    """A notehead in a music score"""

    glyph: Glyph
    "The visual glyph of the notehead"

    notes: List[Note]
    "List of notes being represented by this notehead (typically just one)"

    clef: Clef
    "What clef applies to the note (notehead)"

    staff: StaffVisual
    "What stafflines is the notehead placed onto"

    pitch_position: int
    "Pitch position of the notehead on the stafflines"

    up_stem_attachment_side: Optional[NoteheadSide] = NoteheadSide.right
    """On what side of the notehead should an up-pointing stem be attached.
    Notehead placing algorithm can modify this, e.g. in a dense
    chord where noteheads have to be from both sides of the stem.
    None means there should not ever be such a stem attached."""

    down_stem_attachment_side: Optional[NoteheadSide] = NoteheadSide.left
    """On what side of the notehead should a down-pointing stem be attached.
    Notehead placing algorithm can modify this, e.g. in a dense
    chord where noteheads have to be from both sides of the stem.
    None means there should not ever be such a stem attached."""

    def detach(self):
        """Unlink the notehead from the scene"""
        self.glyph.detach()
        self.notes = []

    @classmethod
    def of_note(cls, note: Note):
        return cls.of(note, lambda n: n.notes)



from ..semantic.StemValue import StemValue
from ..semantic.BeamedGroup import BeamedGroup
from ..AffineSpace import AffineSpace
from smashcima.geometry.Point import Point
from ..SceneObject import SceneObject


class BeamCoordinateSystem(SceneObject):
    """Defines the placement and slope of a beamed group. All values are in the
    paper coordinate system (paper space)."""

    def __init__(
        self,
        beamed_group: BeamedGroup,
        page_space: AffineSpace,
        k: float,
        q: float,
        beam_spacing: float
    ):
        """Initialize the coordinate system as a linear function over the
        paper space.
        
        :param beamed_group: The semantic beamed group that is described by this.
        :param space: The paper space in which all the computations occur.
        :param k: The slope of the line (rise over run).
        :param q: The value of Y at X=0.
        :param beam_spacing: Separation between beams in millimeters.
        """
        super().__init__()
        
        self.beamed_group = beamed_group
        self.page_space = page_space
        self.k = k
        self.q = q
        self.beam_spacing = beam_spacing
    
    @classmethod
    def of_beamed_group(cls, beamed_group: BeamedGroup):
        return cls.of(beamed_group, lambda bcs: bcs.beamed_group)
    
    def __call__(
        self,
        x: float,
        beam_number: int = 1,
        stem_value: StemValue = StemValue.up
    ) -> float:
        """The coordinate system acts like a linear function over the paper
        space, so that when you give it X coordinate, it will give you the Y
        vertical coordinate of the main beam (unless you also specify which
        beam you are interested in, and from which direction is the stem
        comming in)."""
        assert beam_number >= 1, type(beam_number) is int

        beam_index = (beam_number - 1)
        
        if stem_value == StemValue.down:
            beam_index *= -1

        return self.k * x + self.q + (beam_index * self.beam_spacing)
    
    def point(
        self,
        x: float,
        beam_number: int = 1,
        stem_value: StemValue = StemValue.up
    ):
        """Invokes the linear function and instead of returning just Y,
        it returns the whole [X, Y] point"""
        return Point(x, self(x, beam_number, stem_value))



from dataclasses import dataclass

from .Region import Region


@dataclass
class LabeledRegion(Region):
    """Scene region with a classification label.

    Represents a region in the scene that could be recognized in
    an object detection task.

    It is defined by a set of polygons, because a polygon can undergo affine
    transformations and remain a polygon (unlike a bounding box).
    Other lower-detail regions can be derived from the set of polygons easily
    (e.g. convex hull, tight rectangle, bounding box).
    """

    label: str
    """Label for the object classification task"""

    def __post_init__(self):
        assert type(self.label) is str, "Label must be a string"



import numpy as np
import random
import heapq


class Quilter:
    """Service that can be used to quilt small textures up to large dimensions"""
    def __init__(
        self,
        rng: random.Random,
        source_texture_block_size_ratio=1/3,
        overlap_ratio=1/10
    ):
        self.rng = rng
        "Random number generator used for patch sampling"

        self.source_texture_block_size_ratio = source_texture_block_size_ratio
        """How large blocks should be cut from the source texture as a ratio
        of the source texture size"""

        self.overlap_ratio = overlap_ratio
        "How much do the sampled blocks overlap when quilting them together"

    def quilt_texture_to_dimensions(
        self,
        source_texture: np.ndarray,
        target_width_px: int,
        target_height_px: int
    ) -> np.ndarray:
        """
        Given a source texture, it quilts the texture up to the desired size.
        """
        # how large square blocks (in pixels) do we cut from the source texture
        block_size_px = int(
            (min(source_texture.shape[0], source_texture.shape[1]) - 1) \
            * self.source_texture_block_size_ratio
        )

        # how much will the blocks overlap (in pixels)
        overlap_px = int(block_size_px * self.overlap_ratio)
        if overlap_px == 0:
            raise Exception("Quilting block overlap is 0px.")
        
        # how many blocks are there going to be in each direction
        # (make the resulting texture slightly larger and then crop it)
        num_blockHigh = target_height_px // (block_size_px - overlap_px) + 2
        num_blockWide = target_width_px // (block_size_px - overlap_px) + 2

        # prepare the resulting texture (the canvas)
        h = (num_blockHigh * block_size_px) - (num_blockHigh - 1) * overlap_px
        w = (num_blockWide * block_size_px) - (num_blockWide - 1) * overlap_px
        canvas = np.zeros(
            shape=(h, w, source_texture.shape[2]),
            dtype=np.uint8
        )

        # place all blocks
        for i in range(num_blockHigh):
            for j in range(num_blockWide):
                y = i * (block_size_px - overlap_px)
                x = j * (block_size_px - overlap_px)

                # Fast sample and clean cut.
                patch = self.random_patch(source_texture, block_size_px)
                patch = self.overlap_canvas_over_patch(
                    patch, overlap_px, canvas, y, x
                )
                
                canvas[y:y+block_size_px, x:x+block_size_px] = patch
        
        # crop out the desired size randomly from the slightly larger canvas
        over_height = canvas.shape[0] - target_height_px
        over_width = canvas.shape[1] - target_width_px
        offset_height = self.rng.randint(0, over_height)
        offset_width = self.rng.randint(0, over_width)
        return canvas[
            offset_height:target_height_px + offset_height,
            offset_width:target_width_px + offset_width
        ]
    
    def overlap_canvas_over_patch(
        self,
        patch: np.ndarray, 
        overlap: int, 
        canvas: np.ndarray, 
        y: int, 
        x: int
    ) -> np.ndarray:
        patch = patch.copy()
        dy, dx, _ = patch.shape
        minCut = np.zeros_like(patch, dtype=bool)

        if x > 0:
            left = (patch[:, :overlap] - canvas[y:y+dy, x:x+overlap]) / 255.0
            leftL2 = np.sum(left**2, axis=2)
            for i, j in enumerate(self.dijkstra_path(leftL2)):
                minCut[i, :j] = True

        if y > 0:
            up = (patch[:overlap, :] - canvas[y:y+overlap, x:x+dx]) / 255.0
            upL2 = np.sum(up**2, axis=2)
            for j, i in enumerate(self.dijkstra_path(upL2.T)):
                minCut[:i, j] = True

        np.copyto(patch, canvas[y:y+dy, x:x+dx], where=minCut)

        return patch

    def dijkstra_path(
        self,
        errors: np.ndarray
    ):
        """
        Dijkstra's algorithm for finding the shortest path in a graph vertically.
        """
        pq = [(error, [i]) for i, error in enumerate(errors[0])]
        heapq.heapify(pq)

        h, w = errors.shape
        seen = set()

        while pq:
            error, path = heapq.heappop(pq)
            curDepth = len(path)
            curIndex = path[-1]

            if curDepth == h:
                return path

            for delta in -1, 0, 1:
                nextIndex = curIndex + delta

                if 0 <= nextIndex < w:
                    if (curDepth, nextIndex) not in seen:
                        cumError = error + errors[curDepth, nextIndex]
                        heapq.heappush(pq, (cumError, path + [nextIndex]))
                        seen.add((curDepth, nextIndex))
    
    def random_patch(self, source_texture: np.ndarray, block_size_px: int):
        """Take a random square block patch from the source texture"""
        h, w, _ = source_texture.shape
        i = np.random.randint(h - block_size_px)
        j = np.random.randint(w - block_size_px)
        return source_texture[i:i+block_size_px, j:j+block_size_px]



from dataclasses import dataclass


@dataclass
class KeySignature:
    """Represents music notation key signature"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/elements/key/
    
    fifths: int
    """Number of flats or sharps of a key signature. Negative values are flats.
    Zero is no accidentals."""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/elements/fifths/

    def __post_init__(self):
        assert self.fifths in list(range(-7, 7+1)), \
            "Key signature fifths must be between -7 and +7"



from ..Glyph import Glyph
from ..SceneObject import SceneObject
from .Notehead import Notehead
from dataclasses import dataclass


@dataclass
class Accidental(SceneObject):
    """Glyph of an accidental"""

    glyph: Glyph
    """The glyph of the accidental"""

    notehead: Notehead
    """The notehead that the accidental belongs to."""

    def detach(self):
        """Unlink the glyph from the scene"""
        self.glyph.detach()
        self.notehead = None



from .StyleDomain import StyleDomain
from smashcima.assets.AssetRepository import AssetRepository
from smashcima.assets.textures.MzkPaperPatches import MzkPaperPatches, Patch
import random
from typing import List


class MzkPaperStyleDomain(StyleDomain):
    """This style domain represents the set of paper texture patches
    that are used to synthesize the paper texture."""

    def __init__(self, assets: AssetRepository, rng: random.Random):
        bundle = assets.resolve_bundle(MzkPaperPatches)
        
        self.all_patches: List[Patch] = bundle.load_patch_index()
        "All paper texture patches available in the bundle"
        
        self.rng = rng
        "The RNG used for randomness"

        assert len(self.all_patches) > 0, "There must be at least one patch"

        # use the first patch as the default
        # (the pick_style will be called before the first sampling)
        self.current_patch: Patch = self.all_patches[0]
        "The patch to be used for the currently synthesized sample"

    def pick_style(self):
        self.current_patch = self.rng.choice(self.all_patches)



from dataclasses import dataclass

from smashcima.geometry import Rectangle

from .AffineSpace import AffineSpace
from .SceneObject import SceneObject


@dataclass
class ViewBox(SceneObject):
    """Viewport into the scene, used for framing exports."""

    space: AffineSpace
    """The space in which the view is placed"""

    rectangle: Rectangle
    """The rectangular position of the view in its space"""



from enum import Enum


class NoteheadSide(str, Enum):
    """Represents the left or right side of a notehead. This type is used to
    codify on what side of the notehead a given stem should be attached."""
    left = "left"
    right = "right"



import copy
from pathlib import Path
from typing import List, Optional, Union

import numpy as np

from smashcima.geometry import Vector2
from smashcima.loading import load_score
from smashcima.exporting import BitmapRenderer
from smashcima.scene import AffineSpace, Page, Scene, Score
from smashcima.synthesis import (BeamStemSynthesizer, ColumnLayoutSynthesizer,
                                 LineSynthesizer, MuscimaPPGlyphSynthesizer,
                                 MuscimaPPLineSynthesizer,
                                 MuscimaPPStyleDomain, MzkPaperStyleDomain,
                                 MzkQuiltingPaperSynthesizer,
                                 NaiveLineSynthesizer,
                                 NaiveStafflinesSynthesizer, PaperSynthesizer,
                                 SimplePageSynthesizer,
                                 SolidColorPaperSynthesizer,
                                 StafflinesSynthesizer,
                                 GlyphSynthesizer)
from smashcima.synthesis.style.MzkPaperStyleDomain import Patch

from .Model import Model


class BaseHandwrittenScene(Scene):
    """Scene synthesized by the `BaseHandwrittenModel`"""
    def __init__(
        self,
        root_space: AffineSpace,
        score: Score,
        mpp_writer: int,
        mzk_background_patch: Patch,
        pages: List[Page],
        renderer: BitmapRenderer
    ):
        super().__init__(root_space)
        
        self.score = score
        """The semantic score based on which the scene was synthesized"""

        self.mpp_writer = mpp_writer
        """The MUSCIMA++ writer number that was used for this scene"""

        self.mzk_background_patch = mzk_background_patch
        """The MZK texture patch used for the background paper"""

        self.pages = pages
        """All the pages of music that were synthesized"""

        self.renderer = renderer
        """The renderer to be used for page rasterization"""

        # add to the list of scene objects
        self.add_many([score, *pages])

    def render(self, page: Page) -> np.ndarray:
        """Renders the bitmap BGRA image of a page"""
        assert page in self.pages, "Given page is not in this scene"
        return self.renderer.render(page.view_box)


class BaseHandwrittenModel(Model[BaseHandwrittenScene]):
    """Synthesizes handwritten pages of music notation.

    This model provides similar functionality as MuseScore when it comes
    to rendering music content. You put in a musical content file
    (say MusicXML) and you get out a scene with a number of pages
    (depending on the music length and system and page breaks)
    and the scene can then be turn into an image and other annotations.
    
    This model acts as a flagship demonstrator for the Smashcima library.
    It serves as an example of what a well-designed Model looks like.
    """

    def register_services(self):
        super().register_services()
        c = self.container
        
        c.type(ColumnLayoutSynthesizer)
        c.type(BeamStemSynthesizer)
        c.interface(StafflinesSynthesizer, NaiveStafflinesSynthesizer)
        c.interface(GlyphSynthesizer, MuscimaPPGlyphSynthesizer)
        c.interface(LineSynthesizer, MuscimaPPLineSynthesizer)
        c.type(SimplePageSynthesizer)
        c.type(MuscimaPPStyleDomain)
        c.type(MzkPaperStyleDomain)
        # c.interface(PaperSynthesizer, SolidColorPaperSynthesizer)
        c.interface(PaperSynthesizer, MzkQuiltingPaperSynthesizer)

    def resolve_services(self):
        super().resolve_services()
        c = self.container

        self.layout_synthesizer = c.resolve(ColumnLayoutSynthesizer)
        self.page_synthesizer = c.resolve(SimplePageSynthesizer)

        self.mpp_style_domain = c.resolve(MuscimaPPStyleDomain)
        self.mzk_paper_style_domain = c.resolve(MzkPaperStyleDomain)
    
    def configure_services(self):
        super().configure_services()
        
        self.styler.register_domain(
            MuscimaPPStyleDomain,
            self.container.resolve(MuscimaPPStyleDomain)
        )
        self.styler.register_domain(
            MzkPaperStyleDomain,
            self.container.resolve(MzkPaperStyleDomain)
        )

    def __call__(
        self,
        file: Union[Path, str, None] = None,
        data: Union[bytes, str, None] = None,
        format: Optional[str] = None,
        score: Optional[Score] = None,
        clone_score: bool = False
    ) -> BaseHandwrittenScene:
        """Synthesizes handwritten pages given a musical content.
        
        The musical content can be provided as a file path, string data,
        or an already parsed Smashcima Score object. The content will
        be placed onto a page until it overflows and then another page is
        added - just like using MuseScore for MXL rendering.

        :param file: Path to a file with musical content
            (e.g. './my_file.musicxml')
        :param data: Musical contents as a bytes or string
            (e.g. MusicXML file contents)
        :param format: In what format is the musical content
            (file suffix, including the period, i.e. '.musicxml')
        :param score: Musical content in the form of an already parsed
            Smashcima Score
        :param clone_score: Should the score be cloned before being embedded
            in the resulting scene
        :returns: The synthesized scene with all the pages
        """

        # NOTE: This method is where input pre-processing should happen
        # and where the state of the model should be prepared for the
        # next synthesis invocation. For example, the Model base class
        # lets the styler pick specific styles here. Similarly, after the
        # synthesis core (the call() method) is invoked, this method is where
        # you can do any post-processing and updates to the model state.
        # For example, the Model base class sets the self.scene property here.

        if score is None:
            score = self.load_score(
                file=file,
                data=data,
                format=format
            )
        elif clone_score:
            score = copy.deepcopy(score)

        return super().__call__(score)

    def load_score(
        self,
        file: Union[Path, str, None] = None,
        data: Union[bytes, str, None] = None,
        format: Optional[str] = None,
    ) -> Score:
        """This method is responsible for loading input annotation files.
        
        Override this method to modify the loading behaviour.
        """
        return load_score(
            file=file,
            data=data,
            format=format
        )

    def call(self, score: Score) -> BaseHandwrittenScene:
        # NOTE: This method is where the synthesis itself happens and
        # the resulting scene is constructed. This method should not modify
        # the state of the model instance, these modifications should happen
        # in the __call__() method instead.

        root_space = AffineSpace()

        # until you run out of music
        # 1. synthesize a page of stafflines
        # 2. fill the page with music
        pages = []
        next_measure_index = 0
        next_page_origin = Vector2(0, 0)
        _PAGE_SPACING = 10 # 1cm
        while next_measure_index < score.measure_count:
            # prepare the next page of music
            page = self.page_synthesizer.synthesize_page(next_page_origin)
            page.space.parent_space = root_space
            pages.append(page)

            next_page_origin += Vector2(
                page.view_box.rectangle.width + _PAGE_SPACING,
                0
            )

            # synthesize music onto the page
            systems = self.layout_synthesizer.fill_page(
                page,
                score,
                start_on_measure=next_measure_index
            )
            next_measure_index = systems[-1].last_measure_index + 1

        # construct the complete scene and return
        return BaseHandwrittenScene(
            root_space=root_space,
            score=score,
            mpp_writer=self.mpp_style_domain.current_writer,
            mzk_background_patch=self.mzk_paper_style_domain.current_patch,
            pages=pages,
            renderer=BitmapRenderer(dpi=300)
        )



import IPython.display
import ipycytoscape
from ..scene.Scene import Scene
from smashcima.scene.SceneObject import SceneObject, Link
from typing import List
import ipywidgets
import IPython
import pprint
from copy import deepcopy


_CYTOSCAPE_STYLE = [
    {
        "selector": "node",
        "css": {
            "content": "data(name)",
            "text-valign": "center",
            "color": "white",
            "text-outline-width": 2,
            "text-outline-color": "data(color)",
            "background-color": "data(color)"
        }
    },
    {
        "selector": "node:parent",
        "css": {
            "background-opacity": 0.333
        }
    },
    {
        "selector": "edge",
        "style": {
            "width": 4,
            "opacity": 0.5,
            "line-color": "data(color)",
            "target-arrow-shape": "triangle",
            "target-arrow-color": "data(color)",
            "curve-style": "bezier"
        }
    }
]


def _scene_object_to_color(scene_object: SceneObject) -> str:
    typename = str(type(scene_object))
    if "smashcima.scene.visual" in typename: return "green"
    if "smashcima.scene.semantic" in typename: return "#11479e"
    return "gray"


def _scene_object_to_node(scene_object: SceneObject) -> ipycytoscape.Node:
    node = ipycytoscape.Node()
    node.data["id"] = str(id(scene_object))
    node.data["name"] = type(scene_object).__name__
    node.data["color"] = _scene_object_to_color(scene_object)
    return node


def _link_to_edge(link: Link) -> ipycytoscape.Edge:
    edge = ipycytoscape.Edge()
    edge.data["source"] = str(id(link.source))
    edge.data["target"] = str(id(link.target))
    edge.data["color"] = _scene_object_to_color(link.source)
    return edge


def _populate_graph_from_scene(graph: ipycytoscape.Graph, scene: Scene):
    nodes: List[ipycytoscape.Node] = []
    edges: List[ipycytoscape.Edge] = []

    for obj in scene.objects.values():
        nodes.append(_scene_object_to_node(obj))
        for outlink in obj.outlinks:
            edges.append(_link_to_edge(outlink))
    
    graph.add_nodes(nodes)
    graph.add_edges(edges)


def display_scene_graph(scene: Scene):
    """Displays an interactive scene graph jupyter widget"""
    
    # === create widgets ===
    
    widget = ipycytoscape.CytoscapeWidget()
    out = ipywidgets.Output()

    # === setup cytoscape graph ===

    _populate_graph_from_scene(widget.graph, scene)
    
    widget.set_layout(name="dagre", nodeSpacing=10, edgeLengthVal=10)
    widget.set_style(_CYTOSCAPE_STYLE)

    # === handle interactions ===

    with out:
        print("Click on node to preview it.")

    def handle_node_click(node: dict):
        with out:
            out.clear_output()
            obj = scene.objects[int(node["data"]["id"])]
            print(pprint.pformat(obj))

    widget.on("node", "click", handle_node_click)

    # === display in jupyter ===

    IPython.display.display(widget)
    IPython.display.display(out)


def display_scene_object_graph(
    obj: SceneObject,
    recurse_via_inlinks=True,
    discard_object_inlinks=True
):
    """Displays an interactive scene graph for a single scene object"""
    # isolate the given object from the graph by making a copy
    # and removing all inlinks
    obj_clone = deepcopy(obj)
    if discard_object_inlinks:
        obj_clone.inlinks = []
    scene = Scene()
    scene.objects.clear() # remove the root affine space
    scene.add(obj_clone, recurse_via_inlinks=recurse_via_inlinks)
    display_scene_graph(scene)



import cv2
import numpy as np

from smashcima.geometry import Point, Rectangle, Transform, Vector2
from smashcima.scene import Glyph, LineGlyph, ScenePoint, Sprite, AffineSpace

from ..LineSynthesizer import LineSynthesizer


class NaiveLineSynthesizer(LineSynthesizer):
    def __init__(self, color=(0, 0, 0, 255), line_width=0.5, dpi=300):
        self.color = color
        "Color of the lines"
        
        self.line_width = line_width
        "The width of the synthesized naive line"

        self.dpi = dpi
        "DPI at which to synthesize the line sprite"

    def create_glyph(self, label: str, delta: Vector2) -> LineGlyph:
        length = delta.magnitude

        # create the glyph space
        space = AffineSpace()

        # create a sprite, lying horizontally, centered on origin
        sprite = Sprite.rectangle(
            space,
            Rectangle(
                x=-length/2,
                y=-self.line_width/2,
                width=length,
                height=self.line_width
            ),
            fill_color=self.color,
            dpi=self.dpi
        )

        # create the two points
        start_point = ScenePoint(
            point=Point(-length/2, 0),
            space=space
        )
        end_point = ScenePoint(
            point=Point(length/2, 0),
            space=space
        )

        # put it all together into a line glyph
        return LineGlyph(
            space=space,
            region=Glyph.build_region_from_sprites_alpha_channel(
                label=label,
                sprites=[sprite]
            ),
            sprites=[sprite],
            start_point=start_point,
            end_point=end_point
        )



import abc


class StyleDomain(abc.ABC):
    """Style domain represents one aspect along side of which the style
    of a synthesized data sample can be affected. For example, the choice
    of a background patch to use for background synthesis would be a style
    dimension that can be captured by a style domain. Where the domain are all
    the possible background patches to choose from. Similarly, the specific
    writer of a handwritten glyph dataset in the context of all of its writers
    would represent a style domain."""
    
    def pick_style(self):
        """This method is called by the Styler before a new sample is
        synthesized. This method should pick a concrete value out of the
        described domain of values."""
        raise NotImplementedError



from dataclasses import dataclass, field
from ..SceneObject import SceneObject
from ..ViewBox import ViewBox
from ..AffineSpace import AffineSpace
from .StaffVisual import StaffVisual
from typing import List


@dataclass
class Page(SceneObject):
    """Represents the visual page of music that is synthesized.
    There may be more than one page in a scene."""

    space: AffineSpace
    "Affine space of the page, origin in the top left corner."

    view_box: ViewBox
    "The view box that renders the page, in the global space"

    staves: List[StaffVisual] = field(default_factory=list)
    "Stafflines on the page, sorted top to bottom"



import abc
from typing import TypeVar

from .AffineSpace import AffineSpace
from .SceneObject import SceneObject

T = TypeVar("T", bound="AffineSpaceVisitor")


class AffineSpaceVisitor(abc.ABC):
    """Base class for walking through the scene hierarchy"""
    def __init__(self, space: AffineSpace):
        self.space = space

    def run(self):
        """Executes the tree visiting algorithm"""
        # IMPORTANT: Iterate in the order in which inlinks are listed!
        for link in self.space.inlinks:
            if isinstance(link.source, AffineSpace):
                sub_visitor = self.create_sub_visitor(link.source)
                sub_visitor.run()
                self.accept_sub_visitor(sub_visitor)
            else:
                self.visit_scene_object(link.source)
    
    @abc.abstractmethod
    def create_sub_visitor(self: T, sub_space: AffineSpace) -> T:
        """Creates the visitor instance for a sub space"""
        raise NotImplementedError

    @abc.abstractmethod
    def accept_sub_visitor(self: T, sub_visitor: T):
        """Once sub space visiting finished, incorporate its results"""
        raise NotImplementedError

    @abc.abstractmethod
    def visit_scene_object(self, obj: SceneObject):
        """Handle scene objects that are children but are not affine spaces"""
        raise NotImplementedError



from enum import Enum


class ClefSign(str, Enum):
    """The appearance of the clef sign (G, F, C)"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/data-types/clef-sign/
    
    G = "G"
    F = "F"
    C = "C"
    percussion = "percussion"
    TAB = "TAB"
    jianpu = "jianpu"



from dataclasses import dataclass
from .Durable import Durable
from .Pitch import Pitch
from typing import Optional


@dataclass
class RestSemantic(Durable):
    display_pitch: Optional[Pitch]
    """Display pitch controls where the rest is placed vertically.
    If none, default placement should be used. For more information
    on the positioning, see the RestGlyph class."""



from dataclasses import dataclass
from typing import List

from smashcima.geometry import Contours

from .AffineSpace import AffineSpace
from .Glyph import Glyph
from .LabeledRegion import LabeledRegion
from .Sprite import Sprite


@dataclass
class ComposedGlyph(Glyph):
    """A group of glyphs that make up a larger logical glyph.

    Sometimes a glyph can be represented as a whole, but can also be rendered
    as a group of smaller glyphs. Always forcing the rendering from parts
    might be detremental, because we might not have the assets to synthesize
    from. But if we allow rendering larger glyphs, they may be:
    1) a ligature which can be rendered as a whole thing or nothing
    2) a composite which can be split up into parts
    If we allow rendering the first part, we loose the subparts, and that's
    a compromise we have to make. However with the second part, we can preserve
    the information about the composition.

    Sub-glyphs should have their parent space linked to this glyph's space.
    
    Sprites in this glyph MUST be the union of sprites of subglyphs,
    so that all the external users can access this property.
    You can do so by calling aggregate_sprites() once the glyph is built.
    
    Methods here have to be overriden to make this collection-glyph behave
    properly.

    This is not really a ligature, as a ligature is something that cannot be
    more subdivided. But when it can be, it will be represented by this class.
    So this class is like a "decomposable" ligature.
    """
    
    sub_glyphs: List[Glyph]
    """Glyphs that make up this larger glyph"""

    def detach(self):
        """Unlink the glyph from the scene"""
        super().detach()
        for g in self.sub_glyphs:
            g.detach()
        self.sub_glyphs = []

    @staticmethod
    def build(label: str, sub_glyphs: List[Glyph]) -> "ComposedGlyph":
        """Builds a composed glyph from a collection of its sub-glyphs.

        Given sub-glyphs must not be attached to the scene yet, meaning their
        space must have no parent space. Their relative position within the
        composed glyph's space must be set by configuring their transform
        before (or after) they are passed into this fuction.

        These glyphs will be attached to the newly constructed composed
        glyph's space, their sprites will be referenced directly and their
        regions will be used to compute the unified region for the composed
        glyph.

        :param label: Classification label for the composed glyph
        :param sub_glyphs: Glyphs that will be included in the composed glyph
        """
        space = AffineSpace()

        assert all(g.space.parent_space is None for g in sub_glyphs), \
            "Given sub-glyphs must NOT be attached to the scene"
        
        # attach sub glyphs under the new root space
        for g in sub_glyphs:
            g.space.parent_space = space
        
        # gather sprites
        sprites: List[Sprite] = [s for g in sub_glyphs for s in g.sprites]

        # gather contours
        contours = Contours([
            c for g in sub_glyphs
            for c in g.region.get_contours_in_space(space).polygons
        ])

        return ComposedGlyph(
            space=space,
            region=LabeledRegion(
                space=space,
                contours=contours,
                label=label
            ),
            sprites=sprites,
            sub_glyphs=sub_glyphs
        )

    def place_debug_overlay(self) -> List[Sprite]:
        overlay: List[Sprite] = []

        overlay += super().place_debug_overlay()
        for g in self.sub_glyphs:
            overlay += g.place_debug_overlay()
        
        return overlay



from enum import Enum


class AccidentalValue(str, Enum):
    """Presence of an accidental for a note"""
    # https://www.w3.org/2021/06/musicxml40/musicxml-reference/data-types/accidental-value/
    sharp = "sharp"
    natural = "natural"
    flat = "flat"
    doubleSharp = "double-sharp"

    # sharpSharp = "sharp-sharp"
    # flatFlat = "flat-flat"
    # naturalSharp = "natural-sharp"
    # naturalFlat = "natural-flat"



from ..SceneObject import SceneObject
from ..LineGlyph import LineGlyph
from ..semantic.Chord import Chord
from ..semantic.BeamedGroup import BeamedGroup
from dataclasses import dataclass
from ..semantic.BeamValue import BeamValue


@dataclass
class BeamHook(SceneObject):
    """One hook glyph in a beamed group (visually)"""

    glyph: LineGlyph
    "The beam hook glyph"

    beamed_group: BeamedGroup
    "The beamed group this beam belongs to"

    chord: Chord
    """The chord in whose stem the hook is placed"""

    beam_number: int
    """Which beam is this beam (1 being the 8th notes beam).
    Uses the MusicXML numbering."""

    hook_type: BeamValue
    """Either forward hook or backward hook. Any other value is invalid."""

    def detach(self):
        """Unlink the glyph from the scene"""
        self.glyph.detach()
        self.beamed_group = None
        self.chord = None



import bisect
import random

from .PackedGlyph import PackedGlyph


class LineList(list):
    """Container that keeps a list of line glyphs and provides their sampling"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        for glyph in self:
            assert isinstance(glyph, PackedGlyph), "Glyphs must be packed"
            assert glyph.line_length is not None, "Glyphs must be line glyphs"
        
        self.sort(key=lambda pg: pg.line_length)
        self.line_lengths = [pg.line_length for pg in self]
    
    def pick_line(
        self,
        target_length: float,
        rng: random.Random,
        percentile_spread=0.1
    ) -> PackedGlyph:
        center = bisect.bisect_left(
            self.line_lengths,
            target_length,
            0,
            len(self)
        )
        target_items = max(int(len(self) * percentile_spread), 2)
        
        # build neighborhood indices
        start = center - target_items // 2 # inclusive
        end = center + target_items // 2 # exclusive
        
        # clamp end
        if end > len(self):
            shift = end - len(self)
            start -= shift
            end -= shift
        
        # clamp start
        if start < 0:
            shift = 0 - start
            start += shift
            end += shift

        # squash end
        if end > len(self):
            end = len(self)
        
        # empty
        if end - start <= 0:
            raise Exception("Cannot sample an empty list")
        
        # sample
        index = rng.randint(start, end - 1)
        return self[index]



from smashcima.scene.LineGlyph import LineGlyph
from smashcima.scene.SmashcimaLabels import SmashcimaLabels
from ..LineSynthesizer import LineSynthesizer
from smashcima.geometry.Vector2 import Vector2
from smashcima.assets.AssetRepository import AssetRepository
from smashcima.assets.glyphs.muscima_pp.MuscimaPPGlyphs import MuscimaPPGlyphs
from smashcima.assets.glyphs.muscima_pp.LineList import LineList
from smashcima.scene.SmuflLabels import SmuflLabels
from smashcima.synthesis.style.MuscimaPPStyleDomain import MuscimaPPStyleDomain
from typing import Dict
import random


_QUERY_TO_MPP_LOOKUP: Dict[str, str] = {
    SmashcimaLabels.ledgerLine.value: SmashcimaLabels.ledgerLine.value,
    SmuflLabels.stem.value: SmuflLabels.stem.value,
    SmashcimaLabels.beam.value: SmashcimaLabels.beam.value,
    SmashcimaLabels.beamHook.value: SmashcimaLabels.beamHook.value,
}


class MuscimaPPLineSynthesizer(LineSynthesizer):
    """
    Synthesizes line glyphs by sampling from the MUSCIMA++ dataset
    """
    def __init__(
        self,
        assets: AssetRepository,
        mpp_style_domain: MuscimaPPStyleDomain,
        rng: random.Random
    ):
        bundle = assets.resolve_bundle(MuscimaPPGlyphs)
        self.symbol_repository = bundle.load_symbol_repository()
        "The symbol repository used for synthesis"

        self.mpp_style_domain = mpp_style_domain
        "Dictates which MUSCIMA++ writer to use for synthesis"
        
        self.rng = rng
        "RNG used for randomization"
    
    def create_glyph(self, label: str, delta: Vector2) -> LineGlyph:
        # translate glyph class
        if label not in _QUERY_TO_MPP_LOOKUP:
            raise Exception("Unsupported glyph label: " + label)
        mpp_label = _QUERY_TO_MPP_LOOKUP[label]

        # select the proper glyph list
        glyphs = self.symbol_repository.glyphs_by_class_and_writer.get(
            (mpp_label, self.mpp_style_domain.current_writer)
        ) or self.symbol_repository.glyphs_by_class.get(mpp_label)
        assert isinstance(glyphs, LineList), \
            f"Got line glyphs without index for {mpp_label}"

        if glyphs is None or len(glyphs) == 0:
            raise Exception(
                f"The glyph class {label} is not present in " + \
                "the symbol repository"
            )

        # pick a random glyph from the list and copy it
        packed_glyph = glyphs.pick_line(delta.magnitude, self.rng)

        # deserialization makes sure we create a new instance here
        glyph = packed_glyph.unpack()
        assert isinstance(glyph, LineGlyph), \
            "The unpacked glyph is not a LineGlyph"

        # ensure the label
        glyph.label = label

        return glyph




from math import ceil
from typing import Tuple

import cv2
import numpy as np

from smashcima.geometry import Quad, Rectangle, Transform, mm_to_px
from smashcima.scene import Sprite, ViewBox

# Alpha compositing via the "over" operator + alpha premultiplication:
# https://en.wikipedia.org/wiki/Alpha_compositing


def _premultiplied_float32_alpha_overlay(canvas: np.ndarray, layer: np.ndarray):
    factor = (1 - layer[:, :, 3:4])
    canvas *= factor
    canvas += layer


def _premultiplied_float32_alpha_overlay_in_window(
    canvas: np.ndarray,
    window: Rectangle,
    layer: np.ndarray
):
    assert int(window.height) == layer.shape[0]
    assert int(window.width) == layer.shape[1]
    top = int(window.top)
    bottom = int(window.bottom)
    left = int(window.left)
    right = int(window.right)
    _premultiplied_float32_alpha_overlay(
        canvas[top:bottom, left:right],
        layer
    )


def _uint8_to_float32(img: np.ndarray) -> np.ndarray:
    img = img.astype(np.float32)
    img /= 255
    return img


def _float32_to_uint8(img: np.ndarray) -> np.ndarray:
    img *= 255
    img = img.astype(np.uint8)
    return img


class BitmapRenderer:
    """Renders a scene into a bitmap RGBA opencv representation"""
    def __init__(
        self,
        dpi: float = 300,
        background_color = (0, 0, 0, 0)
    ):
        self.dpi = float(dpi)
        """DPI at which the scene should be rasterized"""

        self.background_color: Tuple[int, int, int, int] = background_color
        """Color to use for the blank canvas, transparent by default
        (BGRA uint8 format)"""

    def render(self, view_box: ViewBox) -> np.ndarray:
        # bounding box of the canvas in pixel space
        canvas_px_bbox = Rectangle(
            x=0,
            y=0,
            width=ceil(mm_to_px(view_box.rectangle.width, dpi=self.dpi)),
            height=ceil(mm_to_px(view_box.rectangle.height, dpi=self.dpi)),
        )

        # the canvas pixel array in alpha premultiplied float32 format
        canvas = np.zeros(
            shape=(int(canvas_px_bbox.height), int(canvas_px_bbox.width), 4),
            dtype=np.float32
        )

        # fill with background color
        background_color_premultiplied = _uint8_to_float32(cv2.cvtColor(
            np.array([[self.background_color]], dtype=np.uint8),
            cv2.COLOR_RGBA2mRGBA
        ))
        canvas[:, :] = background_color_premultiplied

        # converts from scene millimeter coordinate system
        # to the canvas pixel coordinate system
        scene_to_canvas_transform = (
            Transform.translate(-view_box.rectangle.top_left_corner.vector)
                .then(Transform.scale(mm_to_px(1, dpi=self.dpi)))
        )

        # root scene space
        root_space = view_box.space.get_root()

        # go through all the sprites in the scene
        for (sprite, sprite_transform) in Sprite.traverse_sprites(
            root_space,
            include_pixels_transform=True,
            include_sprite_transform=True,
            include_root_space_transform=False
        ):
            # build up a transform that converts from sprite's local pixel space
            # to canvas global pixel space, while going through the scene space
            to_canvas_transform = (
                sprite_transform # recursive scene hierarchy transforms
                .then(scene_to_canvas_transform)
            )

            # get the window in the canvas that we're going to paint over
            canvas_window: Rectangle = (
                to_canvas_transform.apply_to(
                    Quad.from_rectangle(
                        sprite.pixels_bbox.dilate(1.0) # grow by 1 pixel
                        # dilation is done to accommodate the aliasing blur
                    )
                ) # get the quad of the dilated sprite quad in canvas coordinates
                .bbox() # get the bounding box rectangle
                .snap_grow() # round to integer by growing
                .intersect_with(canvas_px_bbox) # clamp inside of canvas
            )
            to_window_transform = to_canvas_transform.then(
                Transform.translate(-canvas_window.top_left_corner.vector)
            )

            # viewport culling:
            # do not render sprites that have no overlap with the canvas
            if canvas_window.has_no_area:
                continue

            # prepare the sprite bitmap into mRGBA float
            sprite_bitmap = sprite.bitmap
            sprite_bitmap = cv2.cvtColor(sprite_bitmap, cv2.COLOR_RGBA2mRGBA)
            sprite_bitmap = _uint8_to_float32(sprite_bitmap)

            # get the transformed bitmap of the sprite
            new_layer = cv2.warpAffine(
                src=sprite_bitmap,
                M=to_window_transform.matrix,
                dsize=(int(canvas_window.width), int(canvas_window.height)),
                flags=(
                    cv2.INTER_AREA # used for downscaling
                    if to_window_transform.determinant < 1.0
                    else cv2.INTER_LINEAR # used for upscaling
                ),
                borderMode=cv2.BORDER_CONSTANT
            )

            # composit the next layer over the canvas in the window
            _premultiplied_float32_alpha_overlay_in_window(
                canvas, canvas_window, new_layer
            )

        # convert to uint8 RGBA (BGRA actually) and return
        return cv2.cvtColor(
            _float32_to_uint8(canvas),
            cv2.COLOR_mRGBA2RGBA
        )



import abc
import random
from typing import Generic, Optional, TypeVar

from smashcima.assets.AssetRepository import AssetRepository
from smashcima.synthesis.style.Styler import Styler

from .Container import Container


T = TypeVar("T")
"""The scene type the model returns (does NOT need to inherit from `Scene`)"""


class Model(Generic[T], abc.ABC):
    """Base class for a data-generating function used for data synthesis.

    A model is a generating function, that constructs a new scene, whenever
    invoked, based on the constructor and invocation arguments. The goal
    of a model is to wire together and configure synthesizers that together
    construct a new scene, whenever the model is invoked. It differs from a
    synthesizer in that is already comes pre-configured for a singular task
    and requires no configuration to start using. Synthesizer, on the other
    hand, is a LEGO piece that depends on abstract interfaces and must be
    painfully wired together with other synthesizers in order to work
    and it serves some small, abstract task.
    """

    def __init__(self) -> None:
        self.container = Container()
        """IoC container with services used during synthesis"""

        self.scene: Optional[T] = None
        """The scene synthesized during the last invocation of this model"""

        self.register_services()
        self.resolve_services()
        self.configure_services()
    
    def register_services(self):
        """Registers services into the service container.

        Called from the constructor. Override this to customize the behaviour
        of this model at the service level.
        """

        # register the default asset repository into the container,
        # so that synthesizers can resolve asset bundles
        self.container.instance(AssetRepository, AssetRepository.default())

        # register the default RNG to use during randomization
        self.container.instance(random.Random, random.Random())

        # register the styler
        self.container.type(Styler)

    def resolve_services(self) -> None:
        """Defines model fields that hold specific services.
        
        Called from the constructor in order to resolve the specific
        instances of services that will be used during synthesis.
        These services should then be assigned to a well-named model
        fields so that they can be easily accessed during synthesis.
        Override this to resolve additional services from the container.
        """

        self.rng: random.Random = self.container.resolve(random.Random)
        """The RNG that should be used for all synthesis randomness"""

        self.styler: Styler = self.container.resolve(Styler)
        """Controls the style selection for all the synthesizers"""
    
    def configure_services(self):
        """Modifies and configures resolved servies.
        
        Called from the consturctor after services are resolved.
        Here, services should be set-up after their instantiation.
        """
        pass
    
    def __call__(self, *args, **kwargs) -> T:
        """Synthesizes a new scene based on the arguments and returns it.

        Override this to specify what arguments your model expects
        and perform any pre-synthesis and post-synthesis state changes
        to the model instance (e.g. select styles, remember the scene).
        """

        # select the styles used for synthesis of this sample
        self.styler.pick_style()

        # run the synthesis pipeline and build the scene
        self.scene = self.call(*args, **kwargs)

        # return the new scene
        return self.scene

    @abc.abstractmethod
    def call(self, *args, **kwargs) -> T:
        """Implements the synthesis process, returns a new scene.
        
        Arguments passed to this method should be prepared in the __call__()
        method. This method should not modify the state of the model instance,
        it should only read it.
        """
        raise NotImplementedError(
            f"Model {self.__class__.__name__} does not have the `call()` "
            "method implemented."
        )
